/******************************************************************************
 * FILE PURPOSE:    ESWITCH Driver support Source Code
 ******************************************************************************
 * FILE NAME:       eswitch.c
 *
 * DESCRIPTION:     ESWITCH Driver Source
 * 
 * REVISION HISTORY:
 *   19th NOV. 03 - PSP TII  
 *    8th NOV. 03 - PSP TII  

 *
 * (C) Copyright 2003, Texas Instruments, Inc
 *******************************************************************************/
#include <linux/kernel.h> 
#include <linux/module.h> 
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/delay.h>     
#include <linux/spinlock.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <asm/io.h>

#include <linux/skbuff.h>
#include <asm/uaccess.h>

#include <asm/mips-boards/prom.h>
#include <linux/string.h>
#include <linux/config.h>

#include <asm/avalanche/avalanche_map.h>

#include "eswitchHalLx.h"
#include "eswitch.h"
#include "eswitchConfig.h"
 
#define ESWITCH_END_DEBUG 0
#define dbgPrint if (ESWITCH_END_DEBUG) printk
#define errPrint printk
#define SWITCH_CONF 0
#define switchDebug  if (SWITCH_CONF) printk
static char l3_align_array[]       = {0x02, 0x01, 0x00, 0x03};
#define L3_ALIGN(i)                l3_align_array[i]

#define TCI_LENGTH                 2
#define TPID_LENGTH                2
#define CONTROL_802_1Q_SIZE        (TCI_LENGTH + TPID_LENGTH)

unsigned int eswitch_cpu_freq = 0;
static unsigned int DrvPhyStatus = 0;
static struct net_device *eswitch_device = NULL;

char eswitch_version[] = "5.7.3";
static int psp_pkt_cnt = 0;

/*-----------------------------------------------------------------------------
 * Private functions.
 *---------------------------------------------------------------------------*/
static int  eswitch_dev_init (struct net_device *p_dev);
static int  eswitch_open (struct net_device *dev );
static int  eswitch_dev_close (struct net_device *p_dev);
static int  eswitch_dev_tx (struct sk_buff *skb, struct net_device *p_dev);
static struct net_device_stats *eswitch_dev_get_net_stats (struct net_device *dev);
	
static int eswitch_p_timer_init(ESWITCH_PRIVATE_INFO_T*);
static void eswitch_p_tick_timeout(unsigned long p_cb_param);
inline static int eswitch_p_start_timer(struct timer_list *p_timer, unsigned int delay_ticks);
static int  eswitch_p_stop_timer(struct timer_list *p_timer);

static int eswitch_p_setup_driver_params(ESWITCH_PRIVATE_INFO_T*);
static int eswitch_do_ioctl (struct net_device *p_dev, struct ifreq *ifr, int cmd); 
static int eswitch_hal_send_complete(OS_SENDINFO *p_skb );
static int eswitch_p_get_version(char *buf, char **start, off_t offset, int count, int *eof, void *data);
/*Added for debug support, may be removed from final release*/
static int eswitch_p_get_pktcnt(char *buf, char **start, off_t offset, int count, int *eof, void *data);

static BRG_STS eswitch_config (ESWITCH_DRV_HAL_INFO_T *p_drv_hal, ESWITCH_COMMAND command, unsigned int port, int value );
static BRG_STS eswitch_vlan_config (ESWITCH_DRV_HAL_INFO_T *p_drv_hal, ESWITCH_VLAN_COMMAND command, 
		                    unsigned int Vlan_id, unsigned int egr_prt_map, unsigned int untag_prt_map );
static BRG_STS eswitch_prio_map (ESWITCH_DRV_HAL_INFO_T *p_drv_hal, unsigned char port,
   	                         unsigned char prio_in, unsigned char prio_out );

static int eswitch_p_get_version(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len;
	len = sprintf(buf, "Texas Instruments ESWITCH Driver Version %s\n", eswitch_version);
	return len;
}

static int eswitch_p_get_pktcnt(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len;
	len = sprintf(buf, "Pkts received on Host port = %d\n", psp_pkt_cnt);
	return len;
}

static int eswitch_do_ioctl (struct net_device *p_dev, struct ifreq *ifr, int cmd)
{
    ESWITCH_PRIVATE_INFO_T  *p_eswitch_priv = p_dev->priv;
    ESWITCH_DRV_HAL_INFO_T  *p_drv_hal      = p_eswitch_priv->drv_hal;
    SwitchConsCmd           cmd_struct;
    unsigned int            ret_val   = -1;
    char* func_name;     
 
    /*
     * Only SIOCDEVPRIVATE is used as cmd argument and hence, there is no switch statement required. 
     * Function calls are based on cmd_struct.name
     */
    
    if(!(p_dev && p_dev->flags & IFF_UP))
    {
	    printk("The eswitch interface is down and ");
	    return -ENODEV;
    }

    if (copy_from_user(&cmd_struct, (SwitchConsCmd *)(ifr->ifr_data), sizeof(SwitchConsCmd)))
  	   return -EFAULT;

    func_name = cmd_struct.name;
    if(strcmp(func_name, "eswitch_config") == 0)
    {
	ret_val = eswitch_config( p_drv_hal, cmd_struct.cmd, 
			        cmd_struct.CmdType.PortCmd.port,
				cmd_struct.CmdType.PortCmd.value);	
	
    }
    else
    if(strcmp(func_name, "eswitch_vlan_config") == 0)
    {
	switchDebug("in eswitch_vlan_config \n");    	    
	ret_val = eswitch_vlan_config( p_drv_hal, cmd_struct.cmd,
			            cmd_struct.CmdType.VlanCmd.vid,
 		                    cmd_struct.CmdType.VlanCmd.egr_ports,
 				    cmd_struct.CmdType.VlanCmd.untag_ports);
    }
    else
    if(strcmp(func_name, "eswitch_prio_map") == 0)
    {
	switchDebug("in eswitch_prio_map\n");    	    
        ret_val = eswitch_prio_map( p_drv_hal, cmd_struct.CmdType.PrioCmd.port,
		                     cmd_struct.CmdType.PrioCmd.prio_in,  	
				     cmd_struct.CmdType.PrioCmd.prio_out);			       
    }
    switchDebug(" ret_val = %d\n", ret_val);
    cmd_struct.result = ret_val;
    if (copy_to_user( (SwitchConsCmd *)(ifr->ifr_data), &cmd_struct, sizeof(SwitchConsCmd)))
  	   return -EFAULT;
    return ret_val;	
}

inline static int eswitch_p_rx_buf_setup(ESWITCH_RX_CHAN_INFO_T *p_rx_chan)
{
    /* Number of ethernet packets & max pkt length */
    p_rx_chan->chan->tot_buf_size  = p_rx_chan->chan->buffer_size   + 
                                     2*CONTROL_802_1Q_SIZE          +
	                             p_rx_chan->chan->buffer_offset ;

    p_rx_chan->chan->tot_reserve_bytes = L3_ALIGN(p_rx_chan->chan->buffer_offset & 0x3);

    return(0);
}

static int eswitch_p_setup_driver_params( ESWITCH_PRIVATE_INFO_T *p_eswitch_priv )
{
    int i=0;

    ESWITCH_TX_CHAN_INFO_T *p_tx_chan_info = p_eswitch_priv->tx_chan_info;
    ESWITCH_RX_CHAN_INFO_T *p_rx_chan_info = p_eswitch_priv->rx_chan_info;

    /* Timer stuff */
    for(i=0; i < MAX_TIMER; i++)
        p_eswitch_priv->timer[i].function = NULL;

    /* Tx channel related.*/
    p_tx_chan_info->cfg_chan  = 1;
    p_tx_chan_info->opened_chan  = 0;
    /* eswitch presently supports only one transmit channel */
    p_tx_chan_info->chan->state         = CHAN_CLOSE; 
    p_tx_chan_info->chan->num_BD        = CFG_TX_NUM_BUF_DESC;  
    p_tx_chan_info->chan->buffer_size   = MAX_FRAME_SIZE;    
    p_tx_chan_info->chan->buffer_offset = CFG_TX_BUF_OFFSET;
    p_tx_chan_info->chan->service_max   = CFG_TX_NUM_BUF_SERVICE;

    /* Assuming just one rx channel for now, To be verified */
    p_rx_chan_info->cfg_chan            = 1;
    p_rx_chan_info->opened_chan         = 0;
    p_rx_chan_info->chan->state         = CHAN_CLOSE;
    p_rx_chan_info->chan->num_BD        = CFG_RX_NUM_BUF_DESC;
    p_rx_chan_info->chan->buffer_size   = MAX_FRAME_SIZE;
    p_rx_chan_info->chan->buffer_offset = CFG_RX_BUF_OFFSET;
    p_rx_chan_info->chan->service_max   = CFG_RX_NUM_BUF_SERVICE;
    
    p_eswitch_priv->stats->rx_discards = 0;
    p_eswitch_priv->stats->tx_discards = 0;
    return(0);
}

static int eswitch_p_timer_init(ESWITCH_PRIVATE_INFO_T *p_eswitch_priv) 
{
    struct timer_list *p_timer = p_eswitch_priv->timer;
    
    init_timer(p_timer);
    p_timer->expires 	= 0;
    p_timer->data    	= (unsigned long) p_eswitch_priv;
    p_timer->function   = eswitch_p_tick_timeout;
    
    return 0;
}    
		
static int eswitch_p_start_timer(struct timer_list *p_timer, unsigned int delay_ticks)
{
    p_timer->expires        = jiffies + delay_ticks;

    if(p_timer->function)
    {
        add_timer(p_timer);
    }
    return(0);
} 


static int eswitch_p_stop_timer(struct timer_list *p_timer)
{
    del_timer_sync(p_timer);
    return(0);
}

static void eswitch_p_tick_timeout(unsigned long p_cb_param)
{
    ESWITCH_PRIVATE_INFO_T   *p_eswitch_priv = (ESWITCH_PRIVATE_INFO_T*) p_cb_param;
    ESWITCH_DRV_HAL_INFO_T   *p_drv_hal      = p_eswitch_priv->drv_hal; 
    struct timer_list        *p_timer        = p_eswitch_priv->timer;

    if(test_bit(0, &p_eswitch_priv->set_to_close))
    {
        return;
    }
    p_drv_hal->hal_funcs->Tick(p_drv_hal->hal_dev);
    eswitch_p_start_timer(p_timer, p_eswitch_priv->delay_ticks);
}

/*******************************************************************************
*
* eswitch_open - start the device
*
* This function calls BSP functions to connect interrupts and start the
* device running in interrupt mode.
*
* RETURNS: OK or ERROR
*
********************************************************************************/
static int eswitch_open( struct net_device *p_dev ) 
{
    ESWITCH_PRIVATE_INFO_T *p_eswitch_priv = p_dev->priv;
    ESWITCH_DRV_HAL_INFO_T *p_drv_hal      = p_eswitch_priv->drv_hal;
    int ret_val = 0;
    p_eswitch_priv->set_to_close = 0;
    
    dbgPrint("starting eswitch Driver\n");
    MOD_INC_USE_COUNT;
    p_drv_hal->hal_funcs->Control(p_drv_hal->hal_dev, hcMacAddress, SETBIN, p_eswitch_priv->mac_addr);    
   
    /* start the eswitch Device */
    if(eswitch_drv_start(p_eswitch_priv, p_eswitch_priv->tx_chan_info, 
        	         p_eswitch_priv->rx_chan_info, CHAN_SETUP) == -1)
    {
        dbgPrint("failed to start the device.\n");
	MOD_DEC_USE_COUNT;
        ret_val = -1;
    }
    else
    {
        netif_start_queue(p_dev);
        p_eswitch_priv->stats->start_tick = jiffies;
	dbgPrint("Started the network queue for %s.\n", p_dev->name);
    }
    return (ret_val);
}

/***************************************************************
 *	eswitch_exit:
 *
 *	Returns:
 *		Nothing
 *	Parms:
 *		None
 *
 *	Frees the device structs and memory associated with eswitch device
 *	(lists and buffers).  It also ureserves the IO port regions
 *	associated with this device.
 *
 **************************************************************/
static void eswitch_exit(void)
{
    struct net_device       *p_dev;
    ESWITCH_PRIVATE_INFO_T    *p_eswitch_priv;
    p_dev = eswitch_device;
    p_eswitch_priv = (ESWITCH_PRIVATE_INFO_T *) p_dev->priv;
    
    dbgPrint("Unloading %s irq=%2d io=%04x\n",p_dev->name, (int) p_dev->irq, (int) p_dev->base_addr);    
    eswitch_drv_cleanup(p_eswitch_priv->drv_hal);    
    release_mem_region(p_dev->base_addr, p_eswitch_priv->dev_size);
    unregister_netdev(p_dev);
    eswitch_device = NULL;
    kfree(p_eswitch_priv->drv_hal);
    kfree(p_dev);
}

static int eswitch_dev_tx( struct sk_buff *skb, struct net_device *p_dev)
{
    ESWITCH_PRIVATE_INFO_T *p_eswitch_priv  = p_dev->priv;
    ESWITCH_DRV_HAL_INFO_T *p_drv_hal       = p_eswitch_priv->drv_hal;
    int channel                             = 0;
    int ret_val                             = 0;
    FRAGLIST                                send_frag_list[1];

    send_frag_list->len  = skb->len;
    send_frag_list->data = skb->data;

#ifdef ESWITCH_TEST
    xdump(skb->data, skb->len, "send");
#endif
    dbgPrint("Send function invoked\n"); 

    dma_cache_wback_inv((unsigned long)skb->data, skb->len);    

    if(p_drv_hal->hal_funcs->Send( p_drv_hal->hal_dev, send_frag_list, 1, skb->len,
			           skb, eswitch_hal_send_complete, channel) != 0)
    {
        p_eswitch_priv->net_dev_stats.tx_errors++;  
        goto eswitch_dev_tx_error; 
    }
    dbgPrint("ret_val %d\n",ret_val);
    return(ret_val);

eswitch_dev_tx_error: 
    ret_val = -1;
    return (ret_val);
} /*eswitch_dev_tx */

static unsigned char str2hexnum(unsigned char c)
{
        if(c >= '0' && c <= '9')
                return c - '0';
        if(c >= 'a' && c <= 'f')
                return c - 'a' + 10;
        if(c >= 'A' && c <= 'F')
                return c - 'A' + 10;
        return 0; /* foo */
}

static void str2eaddr(unsigned char *ea, unsigned char *str)
{
    int i;
    unsigned char num;
    for(i = 0; i < 6; i++) {
        if((*str == '.') || (*str == ':') || (*str == '-'))
            str++;
        num = str2hexnum(*str++) << 4;
        num |= (str2hexnum(*str++));
        ea[i] = num;
    }
}

/***************************************************************
 *	eswitch_dev_init
 *
 *	Returns:
 *		0 on success, error code otherwise.
 *	Parms:
 *		p_dev	The structure of the device to be
 *			init'ed.
 *
 *	This function completes the initialization of the
 *	device structure and driver.  It reserves the IO
 *	addresses and assignes the device's methods.
 *
 *		
 **************************************************************/
static int eswitch_dev_init (struct net_device *p_dev)     
{
    ESWITCH_PRIVATE_INFO_T    *p_eswitch_priv;
    ESWITCH_DRV_HAL_INFO_T    *p_drv_hal;
    ESWITCH_TX_CHAN_INFO_T    *p_tx_chan_info;
    ESWITCH_RX_CHAN_INFO_T    *p_rx_chan_info;   
    ESWITCH_DRV_STATS_T       *p_drv_stats; 
    ESWITCH_CONFIG            **p_eswitch_config;
    char*                     mac_name  = NULL;
    char*                     mac_addr  = NULL;
    int                       mem_size  = 0;
    unsigned long             net_flags = 0;
    int                       i = 0;
    int                       retVal = -1;
       
    dbgPrint("Loading eswitch Linux Net Driver\n");
    
    mem_size =   sizeof(ESWITCH_DRV_HAL_INFO_T)
	       + sizeof(ESWITCH_TX_CHAN_INFO_T)
	       + sizeof(ESWITCH_RX_CHAN_INFO_T)	       	       
	       + sizeof(ESWITCH_DRV_STATS_T)
	       + sizeof(ESWITCH_CONFIG);
           

    if((p_drv_hal = kmalloc(mem_size,  GFP_KERNEL)) == NULL)
    {
        dbgPrint("Failed to allocate memory.\n");
        return(-1);
    }

    memset((char *)p_drv_hal, 0, mem_size);
    
    /* build the eswitch private object */
    p_eswitch_priv                = p_dev->priv;
    p_eswitch_priv->drv_hal       = p_drv_hal;
    p_eswitch_priv->tx_chan_info  = p_tx_chan_info
                                  = (ESWITCH_TX_CHAN_INFO_T*)((char*)p_drv_hal 
  				    + sizeof(ESWITCH_DRV_HAL_INFO_T));
    p_eswitch_priv->rx_chan_info  = p_rx_chan_info 
                                  = (ESWITCH_RX_CHAN_INFO_T*)((char *)p_tx_chan_info 
				    + sizeof(ESWITCH_TX_CHAN_INFO_T));    
    
    p_eswitch_priv->stats         = p_drv_stats
	                          = (ESWITCH_DRV_STATS_T *)((char *)p_rx_chan_info 
  		                    + sizeof(ESWITCH_RX_CHAN_INFO_T));
    p_eswitch_priv->p_config      = (ESWITCH_CONFIG *)((char *)p_drv_stats
  	                            + sizeof(ESWITCH_DRV_STATS_T));
    p_eswitch_config              = &(p_eswitch_priv->p_config);

    p_drv_hal->owner              = p_eswitch_priv;
    
    mac_name = "HWA_ESWITCH";
    mac_addr = prom_getenv(mac_name);     
    
    if(!mac_addr)
    {
        mac_addr="08.00.28.32.06.02";
        printk("Error getting MAC from enviroment for %s\n",mac_name );
        printk("Using default MAC address: %s\n",mac_addr);
        if(mac_name)
        {
            printk("setenv %s xx.xx.xx.xx.xx.xx\n", mac_name);
            printk("to set mac address\n");
        }
    }

    str2eaddr(p_eswitch_priv->mac_addr, mac_addr);

    for ( i=0; i <= ETH_ALEN; i++ )
    {
        /* This sets the hardware address */
	p_dev->dev_addr[i] = p_eswitch_priv->mac_addr[i]; 
    }

    p_eswitch_priv->set_to_close          = 1;

    /* The hal module is initialized */
    if(eswitch_drv_init_module(p_drv_hal, p_dev, 0) == -1)
    {
	dbgPrint("error: init module failed for eswitch\n" ); 
	goto eswitch_init_module_error;
    }

    get_eswitch_config_info(p_eswitch_config);
    
    if(eswitch_p_setup_driver_params(p_eswitch_priv) == -1)
    {
        errPrint("failed to set driver parameters for %s.\n", p_dev->name);
        goto eswitch_init_device_error;
    }

    eswitch_p_rx_buf_setup(p_rx_chan_info);
    
    if(eswitch_p_timer_init(p_eswitch_priv) == -1)
    {
        errPrint("failed to set timer(s) for %s.\n", p_dev->name);	    
        goto eswitch_timer_init_error;
    }

    p_dev->addr_len           = 6;
    p_dev->open               = &eswitch_open;       /*  i.e. Start Device  */
    p_dev->hard_start_xmit    = &eswitch_dev_tx;
    p_dev->stop               = &eswitch_dev_close;
    p_dev->get_stats          = &eswitch_dev_get_net_stats;
    p_dev->do_ioctl           = &eswitch_do_ioctl;    /* Added for ioctl support */
    
    p_dev->flags |= IFF_BROADCAST | IFF_MULTICAST;
    dbgPrint(" p_dev->flags after net_flags  %d\n", p_dev->flags);  
    p_dev->base_addr = (*p_eswitch_config)->EswitchBase ;
    p_eswitch_priv->dev_size = (*p_eswitch_config)->EswitchSize ;
    request_mem_region(p_dev->base_addr, p_eswitch_priv->dev_size, p_dev->name);
    retVal = 0;
    
    return(retVal);

eswitch_timer_init_error:
eswitch_init_device_error:
    eswitch_drv_cleanup(p_drv_hal);

eswitch_init_module_error:
    kfree(p_drv_hal);
    return (retVal);
}

static struct net_device_stats *eswitch_dev_get_net_stats (struct net_device *p_dev)
{
    ESWITCH_PRIVATE_INFO_T *p_eswitch_priv = (ESWITCH_PRIVATE_INFO_T *) p_dev->priv;

    p_eswitch_priv->net_dev_stats.rx_errors  = 0;
    p_eswitch_priv->net_dev_stats.collisions = 0;
    
    return &p_eswitch_priv->net_dev_stats;
}

static int eswitch_hal_send_complete(OS_SENDINFO *p_skb )
{
    ESWITCH_PRIVATE_INFO_T *p_eswitch_priv = p_skb->dev->priv;		
    
    p_eswitch_priv->net_dev_stats.tx_packets++;
    p_eswitch_priv->net_dev_stats.tx_bytes += p_skb->len;        
    dev_kfree_skb_any(p_skb);
    return(0);
}

/***************************************************************
 *	eswitch_dev_close
 *  
 * 	Returns:
 *		An error code.
 *	Parms:
 *		dev	The device structure of the device to
 *			close.
 *
 *	This function shuts down the adapter.  
 *
 ************************************************************/
static int eswitch_dev_close(struct net_device *p_dev)
{
    ESWITCH_PRIVATE_INFO_T *p_eswitch_priv = p_dev->priv;
    ESWITCH_DRV_HAL_INFO_T *p_drv_hal      = p_eswitch_priv->drv_hal;
    int ret_val = 0;

    /* inform the upper layers. */
    netif_stop_queue(p_dev);

    set_bit(0, &p_eswitch_priv->set_to_close);
   
    /* The driver does not re-schedule the tasklet after kill is called. So, this 
       should take care of the bug in the kernel. */
    tasklet_kill(&p_eswitch_priv->eswitch_isr->tasklet);
    
    if(eswitch_drv_stop(p_drv_hal, p_eswitch_priv->tx_chan_info,
			p_eswitch_priv->rx_chan_info,
			CHAN_TEARDOWN | FREE_BUFFER | BLOCKING | COMPLETE) == -1)
    {
        ret_val = -1;
        dbgPrint("Device failed to be closed.\n");
    }
    else
    {
        /* hope that the HAL closes down the tick timer.*/
        dbgPrint("Device %s Closed.\n", p_dev->name);
        p_eswitch_priv->stats->start_tick = jiffies;
        remove_proc_entry("avalanche/eswitch_ver", NULL);
        remove_proc_entry("avalanche/eswitch_dbg", NULL);
        MOD_DEC_USE_COUNT;
    }    
    return (ret_val);
} /* eswitch_dev_close */

/*******************************************************************************
* 
*       Kernel level receive interrupt handler
*
*       This function calls device-dependent routine to copy actual data from
*       the device into the provided memory buffer. After that, the data buffer
*       is linked with the sk_buff structures and the resulting structure
*       is passed to the upper layer.
*
*
*       RETURNS: Success (0)/ Failure (-1)
********************************************************************************/
int eswitch_hal_receive(OS_DEVICE *p_dev, FRAGLIST *fragList, 
  		        unsigned int FragCount, unsigned int pkt_len, 
		        HAL_RECEIVEINFO *halReceiveInfo, unsigned int mode)
{
    ESWITCH_PRIVATE_INFO_T *p_eswitch_priv = p_dev->priv;
    ESWITCH_DRV_HAL_INFO_T *p_drv_hal      = p_eswitch_priv->drv_hal;
    struct sk_buff         *p_skb          = fragList[0].OsInfo;
    p_skb->len                             = fragList[0].len;	
   
    dma_cache_inv((unsigned long)p_skb->data, fragList[0].len); 

    if (FragCount > 1)
    {
	 int                     length;
	 struct  sk_buff         *p_temp_skb;
         ESWITCH_RX_CHAN_INFO_T  *p_rx_chan_info = p_eswitch_priv->rx_chan_info;	 
	 int                     count;

	 p_skb = dev_alloc_skb(pkt_len + p_rx_chan_info->chan->tot_reserve_bytes);

	 if(p_skb == NULL) 
	 {
		 p_eswitch_priv->net_dev_stats.rx_dropped++;
		 goto eswitch_hal_recv_alloc_failed;
	 }
         p_skb->dev = p_dev;
	 skb_reserve(p_skb, p_rx_chan_info->chan->tot_reserve_bytes);

	 for (count =0; count < FragCount; count++)
	 {
		p_temp_skb = fragList[count].OsInfo;
		length = fragList[count].len;                                
		dma_cache_inv((unsigned long)p_temp_skb->data, length);
		memcpy(skb_put(p_skb, length), p_temp_skb->data, length); /*TS Copy operation*/
    	        dev_kfree_skb_any(p_temp_skb);
	 }
    }    
    skb_trim(p_skb, pkt_len);
    p_skb->protocol = eth_type_trans (p_skb, p_dev);
    netif_rx(p_skb);
    p_eswitch_priv->net_dev_stats.rx_packets++;
    p_eswitch_priv->net_dev_stats.rx_bytes += pkt_len;        
    p_drv_hal->hal_funcs->RxReturn(halReceiveInfo, 1);
    psp_pkt_cnt++; /*Added for debug Support*/
    return 0;

eswitch_hal_recv_alloc_failed:
    while(FragCount)
    {
        FragCount--;
        dev_kfree_skb_any(fragList[FragCount].OsInfo);
    }
    return -1;
}

int eswitch_hal_control(OS_DEVICE *p_dev, const char *key, 
	                const char *action, void *value)
{
    ESWITCH_PRIVATE_INFO_T *p_eswitch_priv = p_dev->priv;
    int ret_val = -1;

    if(key == NULL)
    {
        dbgPrint("Encountered NULL key.\n");
	return (-1);
    }

    if(eswitch_ci_strcmp(key, "Sleep") == 0 && value != NULL)
    {
	    unsigned int clocks_per_tick  = eswitch_cpu_freq/HZ;
	    unsigned int requested_clock = *(unsigned int*)value;
	    unsigned int requested_tick  = 0;
	    if ( requested_clock % clocks_per_tick )
		    requested_tick = requested_clock/clocks_per_tick + 1;
	    else
		    requested_tick = requested_clock/clocks_per_tick;
	    mdelay(requested_tick); 
	    ret_val = 0;
    }
    else if( key == hcStateChange)
    {
	DrvPhyStatus = (int )value;        
	ret_val = 0;
    }
    else if(eswitch_ci_strcmp(key, hcTick) == 0 && action != NULL)
    {
	if(eswitch_ci_strcmp(action, "Set") == 0 && value != NULL)
	{
	    if(*(unsigned int*)value == 0)
            {
                eswitch_p_stop_timer(p_eswitch_priv->timer + TICK_TIMER);
                ret_val = 0;
            }	
	    else
	    {
		    unsigned int clocks_per_tick  = eswitch_cpu_freq/HZ;
		    unsigned int requested_clock = *(unsigned int*)value;
		    unsigned int requested_tick  = 0;
		    if ( requested_clock % clocks_per_tick )
			    requested_tick = requested_clock/clocks_per_tick + 1;
		    else
			    requested_tick = requested_clock/clocks_per_tick;
		    p_eswitch_priv->delay_ticks = requested_tick;
		    ret_val = eswitch_p_start_timer(p_eswitch_priv->timer + TICK_TIMER,
				    p_eswitch_priv->delay_ticks );
	    }
	}
        else if(eswitch_ci_strcmp(action, "Clear") == 0)
        {
            ret_val = eswitch_p_stop_timer(p_eswitch_priv->timer + TICK_TIMER);
        }
        else
          ;
    }
    else if(eswitch_ci_strcmp(key, "SioFlush") == 0)
    {
       dbgPrint("\n"); 
    }
    return(ret_val);
}

static int __init eswitch_dev_probe(void)
{
    struct  net_device	        *p_dev;
    ESWITCH_PRIVATE_INFO_T      *p_eswitch_priv;
    size_t		        dev_size;
    int                         failed;
    int                         ret_val = 0;

    char *cpu_freq_ptr = prom_getenv("cpufrequency");

    if(!cpu_freq_ptr)
    {
        dbgPrint("Could not find cpufrequency env variable for ESWITCH \n");
	eswitch_cpu_freq = 25000000;
    }
    else
    {
        dbgPrint("Using freq from the env variable for ESWITCH.\n");
	eswitch_cpu_freq = simple_strtol(cpu_freq_ptr, NULL, 0);
    }
    
    dev_size = sizeof(struct net_device) + sizeof(ESWITCH_PRIVATE_INFO_T);
    if((p_dev = (struct net_device *) kmalloc(dev_size, GFP_KERNEL)) == NULL)
    {
	    dbgPrint( "Could not allocate memory for device.\n" );
	    ret_val = -ENOMEM;
	    return ret_val;
    }
    
    memset((char *)p_dev, 0, dev_size );
    p_eswitch_priv = (ESWITCH_PRIVATE_INFO_T*)(((char *) p_dev) + sizeof(struct net_device));
    p_dev->priv    = p_eswitch_priv;
    p_eswitch_priv->owner = p_dev;
    ether_setup(p_dev);
    p_dev->init = eswitch_dev_init;
    dev_alloc_name(p_dev, "esw");
    failed = register_netdev(p_dev);
    if (failed)
    {
	    dbgPrint("Could not register device because of reason code %d.\n", failed);
	    kfree(p_dev);
	    return -ENODEV;
    }
    else
    {
	    eswitch_device = p_dev;   
 	    create_proc_read_entry("avalanche/eswitch_ver", 0, NULL, eswitch_p_get_version, NULL);
	    create_proc_read_entry("avalanche/eswitch_dbg", 0, NULL, eswitch_p_get_pktcnt, NULL);
    }
    dbgPrint("Installed eswitch instances.\n");
    return ret_val;
} /* init_module */

void *eswitch_hal_malloc_buffer ( unsigned int size, void* mem_base, unsigned int mem_range,
	                          OS_SETUP *p_os_setup, HAL_RECEIVEINFO *HalReceiveInfo, 
 	                          OS_RECEIVEINFO **osReceiveInfo, OS_DEVICE *p_dev)
{
    ESWITCH_RX_CHAN_INFO_T  *p_rx_chan_info    = (ESWITCH_RX_CHAN_INFO_T *)p_os_setup;
    int                     tot_buf_size       = p_rx_chan_info->chan->tot_buf_size;
    int                     tot_reserve_bytes  = p_rx_chan_info->chan->tot_reserve_bytes;
    struct sk_buff          *p_skb;
    void                    *ret_ptr;

    p_skb = dev_alloc_skb(tot_buf_size);

    if(p_skb == NULL)
    {
        dbgPrint("failed to allocate buffer.\n");
        return (NULL);
    }

    p_skb->dev = p_dev;
    skb_reserve(p_skb, tot_reserve_bytes);

    *osReceiveInfo = p_skb;

    ret_ptr = skb_put(p_skb, p_rx_chan_info->chan->buffer_size);

    return(ret_ptr);
}
/*******************************************************************************
*
*       This function provides the interface for managing port based entity 
*       of eSwitch. 
*
*        SYNTAX: BRG_STS eswitch_config ( ESWITCH_COMMAND command, 
*		                           unsigned int port, int value )
*			       
*        ARGUMENTS:   command         Of the type ESWITCH_COMMAND
*		      port            port number of eSwitch 1,2...
*		      value           Depends on the command type
*		      
*               |-----------------------|--------------------------|
*               |   COMMAND             |   VALUES                 |
*               |-----------------------|--------------------------|
*		|ESWITCH_PHY_OP         |  PORT_CONFIG             |
*		|-----------------------|--------------------------|
* 		|ESWITCH_PORT_OP        |  ENBL_DSBL               |
*		|-----------------------|--------------------------| 	
*   		|ESWITCH_VLAN_INGRESS_OP|  ENBL_DSBL               |
*		|-----------------------|--------------------------|
*		|ESWITCH_DROP_UNTAG_OP  |  ENBL_DSBL               |
*		|-----------------------|--------------------------|
*		|ESWITCH_PORT_VLAN_OP   |  1-4094                  |
*		|-----------------------|--------------------------|
*		|ESWITCH_STP_PORT_OP    |  STP_STATE               |
*		|-----------------------|--------------------------|		
*
*	
*	RETURN TYPE: BRG_OK,  for Success
*	             BRG_ERR, for Failure
*	             
********************************************************************************/
static BRG_STS eswitch_config ( ESWITCH_DRV_HAL_INFO_T *p_drv_hal, ESWITCH_COMMAND command, unsigned int port, int value )
{
	int speed_and_mode = 0;
	unsigned int PortLink = 0;
	unsigned int PortSpeed = 0;
	unsigned int PortDuplex = 0;

	switchDebug("command %d, port %d , value %d\n", command, port, value);


	if( (port == 0) || (port > 3))
	{
		switchDebug( "Please enter the port number correctly\n");
		return BRG_ERR;
	}
	
	switch (command)
	{
          case ESWITCH_PHY_OP: /* Configures the phy for duplex mode and speed */
		  
		if ((value < eswitch_auto_ngt) || (value > eswitch_ten_full))
		{
			port --;
			if (port == 2)
			{
			    switchDebug ( "Port 3 is an internal port and is not connected to PHY\n");
			    return BRG_OK;					
			}				
			PortLink   = (DrvPhyStatus >> (port << 3))& ESWITCH_PORT_LINK;
			PortSpeed  = (DrvPhyStatus >> (port << 3))& ESWITCH_PORT_SPEED;
			PortDuplex = (DrvPhyStatus >> (port << 3))& ESWITCH_PORT_DUPLEX;			

			switchDebug("PortLink = %d, PortSpeed = %d, PortDuplex =%d\n",PortLink, PortSpeed, PortDuplex);

			if(PortLink)
			{
				switchDebug ( "Port %d Link UP ", (port+1)) ;
			}
			else
                        {   
                           	switchDebug ( "Port %d Link DOWN ", (port+1)) ;
                                return DrvPhyStatus;
                        }			
			if(PortSpeed)
			{
				switchDebug ( "with 100 Mbps ") ;
			}
			else
			{
                           	switchDebug ( "with 10 Mbps ") ;
			}
			if(PortDuplex)
			{
				switchDebug ( "FULL DUPLEX\n") ;
			}
			else
			{
                           	switchDebug ( "HALF DUPLEX\n") ;
			}
			return DrvPhyStatus;	
	
		}
		switch(value)
		{
			case eswitch_auto_ngt:
				speed_and_mode = SNWAY_AUTO |SNWAY_FD100 | SNWAY_HD100 | SNWAY_FD10 | SNWAY_HD10;
				switchDebug(" port %d  configuration parameter auto negotiation", port);
				break;
				
                        case eswitch_hun_half:
				speed_and_mode = SNWAY_HD100;
				switchDebug(" port %d configuration parameter 100 MBPS and Half Duplex Mode", port);				
				break;
				
                        case eswitch_hun_full:
				speed_and_mode = SNWAY_FD100;
				switchDebug(" port %d configuration parameter 100 MBPS and Full Duplex Mode", port);
				break;
				
			case eswitch_ten_half:	
				speed_and_mode = SNWAY_HD10;
				switchDebug("port %d configuration parameter 10 MBPS and Half Duplex Mode", port);
				break;
				
			case eswitch_ten_full:	
				speed_and_mode = SNWAY_FD10;
				switchDebug(" port %d configuration parameter 10 MBPS and Full Duplex Mode", port);
				break;
				
                        default:
                                switchDebug(" duplex and speed enumerations are not correct \n");
				return BRG_ERR;				
				
		}				
                switch(port)
		{
			case 1:
				if(p_drv_hal->hal_funcs->Control(p_drv_hal->hal_dev, hcMdioPhyMode0, SETBIN, speed_and_mode) != 0)
				{
					switchDebug(" can't be configured\n");           
					return BRG_ERR;
				}
				switchDebug(" configured successfully\n");
				return BRG_OK;
				
                        case 2:
				if(p_drv_hal->hal_funcs->Control(p_drv_hal->hal_dev, hcMdioPhyMode1, SETBIN, speed_and_mode) != 0)
				{
					switchDebug(" can't be configured\n");           
					return BRG_ERR;
				}
				switchDebug(" configured successfully\n");
				return BRG_OK;
				
			default:
				switchDebug(" port enumerations are not correct \n");
				return BRG_ERR;				

		}	          
		break;

	  case ESWITCH_PORT_OP: /* Enable/Disables port */

		  port --;
		  if (value == e_Enable)
		  {
         		if(p_drv_hal->hal_funcs->HalFuncExt->EswitchEnablePort(p_drv_hal->hal_dev, port) != 0)
			{
				switchDebug("Failed to enable the port %d for switch operation \n", (++port));
				return BRG_ERR;				
			}
			switchDebug("port %d successfully enabled\n", (++port));
			return BRG_OK;
		  }
		  else if (value == e_Disable)
		  {
			if(p_drv_hal->hal_funcs->HalFuncExt->EswitchDisablePort(p_drv_hal->hal_dev, port) != 0)
			{
				switchDebug("failed to disable the port %d for switch operation \n", (++port));
				return BRG_ERR;				
			}
			switchDebug("port %d successfully disabled\n", (++port) );
			return BRG_OK;
		  }
		  else
			switchDebug ("Enter the correct Enable/Disable value for port operation\n");
		        return BRG_ERR;
		  break;

	  case ESWITCH_VLAN_INGRESS_OP:/* Enable/Disables port ingress checking feature */

                  port--;
		  if (value == e_Enable)
		  {
         		if(p_drv_hal->hal_funcs->HalFuncExt->EswitchEnableVIDIngress(p_drv_hal->hal_dev, port) != 0)
			{
				switchDebug("Failed to enable the VLAN ingress checking for the port %d\n", (++port));
				return BRG_ERR;				
			}
			switchDebug("port %d's INGRESS checking is successfully enabled\n", (++port));
			return BRG_OK;
		  }
		  else if (value == e_Disable)
		  {
			if(p_drv_hal->hal_funcs->HalFuncExt->EswitchDisableVIDIngress(p_drv_hal->hal_dev, port) != 0)
			{
				switchDebug("Failed to disable the VLAN ingress checking for the port %d\n", (++port));
				return BRG_ERR;				
			}
			switchDebug("port %d's INGRESS checking is successfully disabled\n", (++port));			
			return BRG_OK;
		  }
		  else
			switchDebug ("Enter the correct Enable/Disable value for INGRESS operation\n");
		        return BRG_ERR;
		  break;

	  case ESWITCH_DROP_UNTAG_OP: /* Enable/Disables port drop untag packet feature */

		  port--;
		  if (value == e_Enable)
		  {
         		if(p_drv_hal->hal_funcs->HalFuncExt->EswitchEnableDropUnTaggedIngress(p_drv_hal->hal_dev, port) != 0)
			{
				switchDebug("Failed to Enable the Tag/Untag checking for the port %d\n", (++port));
				return BRG_ERR;				
			}
			switchDebug("port %d is successfully enabled for Tag/Untag checking\n", (++port));
			return BRG_OK;
		  }
		  else if (value == e_Disable)
		  {
	  		if(p_drv_hal->hal_funcs->HalFuncExt->EswitchDisableDropUnTaggedIngress(p_drv_hal->hal_dev, port) != 0)
			{
				switchDebug("Failed to disable the Tag/Untag checking for the port %d\n", (++port));
				return BRG_ERR;				
			}
			switchDebug("port %d is successfully disabled for Tag/Untag checking\n", (++port));
			return BRG_OK;
		  }
		  else
			switchDebug ("Enter the correct Enable/Disable value for Tag/Untag operation\n");
		        return BRG_ERR;
		  break;

	  case ESWITCH_PORT_VLAN_OP: /* Assigns the Port VID to the port */

		port--;
		if (!((value > 0) && (value < 4095)))
		{
			switchDebug ("Please enter the appropriate port vlan Identifier\n");
			return BRG_ERR;
		}
 		if(p_drv_hal->hal_funcs->HalFuncExt->EswitchSetPortVlanId (p_drv_hal->hal_dev, port, value) != 0)
		{
			switchDebug("Failed to assign port VLAN Id. to the port %d\n", (++port));
			return BRG_ERR;				
		}
		switchDebug("Port VLAN ID %d is successfully added to the port %d \n", value, (++port));
		return BRG_OK;

	  case ESWITCH_MODE: /* Enable/Disable Layer 2 protocol MODE */
	
		if (value == e_Enable)
		{
			if(p_drv_hal->hal_funcs->HalFuncExt->EswitchEnableMode(p_drv_hal->hal_dev, value) != 0)
			{
				switchDebug("Failed to Enable the Mode %d\n", value);
				return BRG_ERR;				
			}
			switchDebug("Successfully Enabled the Mode for the switch\n");
			return BRG_OK;
		}
		else if (value == e_Disable)
		{
			if(p_drv_hal->hal_funcs->HalFuncExt->EswitchDisableMode(p_drv_hal->hal_dev, value) != 0)
			{
				switchDebug("Failed to disable the Mode %d\n", value);
				return BRG_ERR;				
			}
			switchDebug("Successfully disabled the mode for the switch\n");
			return BRG_OK;
		}
		else
			switchDebug ("Enter the correct Enable/Disable value for setting the mode\n");
		        return BRG_ERR; 
		break;
		
	  case ESWITCH_PORT_STP_OP: /* Sets the STP PORT STATE to STP MIB defined port state */
		
   	         port--;

		 if ((value == PORT_BLOCK) || (value == PORT_LEARN) || (value == PORT_FWD))
		 {
			 if(p_drv_hal->hal_funcs->HalFuncExt->EswitchStpPortState (p_drv_hal->hal_dev, port, value) != 0)
			 {
				 switchDebug("Failed to set STP port state for port %d\n",(++port) );
				 return BRG_ERR;				
			 }
			 switchDebug("Successfully STP port state is set for port %d\n", (++port));
			 return BRG_OK;
		 }
		 else
			switchDebug ("Enter the correct STP port state value\n");
		        return BRG_ERR;
		 break;

	  default:
		  break;
	}
	return BRG_ERR;	
}

/*******************************************************************************
*
*       This function provides the interface for managing VLAN of eSwitch. 
*
*        SYNTAX: BRG_STS eswitch_vlan_config (ESWITCH_VLAN_COMMAND command,
*					       unsigned int Vlan_id,
*					       unsigned int egr_prt_map,
*					       unsigned int untag_prt_map )
*	       
*        ARGUMENTS:   command         ESWITCH_VLAN_ADD
*		                      ESWITCH_VLAN_DEL
*		      Vlan_id         dot1Q VLAN identifier
*		      egr_prt_map     Egress ports for VLAN
*		      untag_prt_map   Untagged ports for VLAN		      
*
*	
*	RETURN TYPE: BRG_OK,  for Success
*	             BRG_ERR, for Failure
*	             
*******************************************************************************/

static BRG_STS eswitch_vlan_config (ESWITCH_DRV_HAL_INFO_T *p_drv_hal, ESWITCH_VLAN_COMMAND command,
	   	                    unsigned int Vlan_id, unsigned int egr_prt_map, unsigned int untag_prt_map )
{
	int port = 0;

	if (!((Vlan_id > 0) && ( Vlan_id < 4095)))
	{
		switchDebug("Vlan Identifier is out of range 1-4094\n");
		return BRG_ERR;
	}		
      
	switch (command)
	{
		case ESWITCH_VLAN_ADD:
			if (!((egr_prt_map <=7) && (untag_prt_map <=7)))
			{
				switchDebug("Please enter the egress and untagged member ports appropriately\n");
				return BRG_ERR;
			}
			
			 switchDebug("VLAN %d has .... \n ", Vlan_id);
			 for (port = 0; port < 3; port++)
			 {
				 if((egr_prt_map >> port) & 1)
				 {
					 switchDebug("port %d ", (port+1)); 
					 if((untag_prt_map >> port) & 1)
					 	{switchDebug ("UNTAGGED \n");}
					 else
					 	{switchDebug ("TAGGED \n");}
				 }
				 else
				 {
					 if((untag_prt_map >> port) & 1)
					 {
						 {switchDebug ("no such member port %d \n", (port+1));}
						 return BRG_ERR;
					 }
				 }
			}

			if(p_drv_hal->hal_funcs->HalFuncExt->EswitchCamAddVlan(p_drv_hal->hal_dev, Vlan_id,
						                               egr_prt_map, 0, 0, untag_prt_map) != 0)
			{
				switchDebug("Failed to Add VLAN %d to eSwitch\n ", Vlan_id);
				return BRG_ERR;				
			}
			return BRG_OK;
				      
		case ESWITCH_VLAN_DEL:
			if(p_drv_hal->hal_funcs->HalFuncExt->EswitchCamDelVlan(p_drv_hal->hal_dev, Vlan_id) != 0)
			{
				switchDebug("Failed to Delete VLAN %d\n from eSwitch ", Vlan_id);
				return BRG_ERR;				
			}
			switchDebug("Successfully deleted VLAN %d from eSwitch\n", Vlan_id);
			return BRG_OK;

		default:	
			break;
	}
	return BRG_ERR;
}

/*******************************************************************************
*
*       This function provides the interface for managing port priority based entity 
*       for eSwitch. 
*
*        SYNTAX: BRG_STS eswitch_prio_map ( unsigned int port, unsigned int prio_in,
			                     unsigned int prio_out )
*	       
*        ARGUMENTS:   port            Port number ( 1,2 ... )
*		      prio_in         Priority of in-coming packet 
*		      prio_out        Priority of out-going packet 
*
*	
*	RETURN TYPE: BRG_OK,  for Success
*	             BRG_ERR, for Failure
*	             
*******************************************************************************/
static BRG_STS eswitch_prio_map (ESWITCH_DRV_HAL_INFO_T *p_drv_hal, unsigned char port,
		              unsigned char prio_in, unsigned char prio_out )
{
	
	if (!((port >= 1) && (port <= 3)))
	{
		switchDebug ("Please enter the port number appropriately for Priority Mapping \n");
		return BRG_ERR;
	}
	port--;

	if ( (prio_in <=7) && (prio_out <=7))
	{
		if(p_drv_hal->hal_funcs->HalFuncExt->EswitchSetPriorityMapping( p_drv_hal->hal_dev, port,
					                                        prio_in, prio_out, prio_out ) != 0)
		{
			switchDebug("Failed to map the input priority %d to out put priority %d for the port %d\n ",
					                                        prio_in, prio_out, (++port) );
			return BRG_ERR;				
		}
	         switchDebug("Successfully mapping of input priority %d to out put priority %d for the port %d\n is done.",
				                                                prio_in, prio_out, (++port));
		 return BRG_OK;
	}
	return BRG_ERR;
}

module_init(eswitch_dev_probe);
module_exit(eswitch_exit);
