/******************************************************************************
 * FILE PURPOSE:    ESWITCH Linux Network Device Driver Header
 ******************************************************************************
 * FILE NAME:       eswitch.h
 *
 * DESCRIPTION:     ESWITCH Network Device Driver Header
 *
 * REVISION HISTORY:
 * Date            Name                   Details
 *----------------------------------------------------------------------------- 
 * July 14 2003                          Initial Create.
 *
 * (C) Copyright 2003, Texas Instruments, Inc
 *******************************************************************************/
#ifndef ESWITCH_H
#define ESWITCH_H

#include <linux/timer.h>
#include <linux/netdevice.h>
#include <asm/semaphore.h>
#include <linux/ctype.h>
#include <linux/interrupt.h>

#include "eswitchConfig.h"

/*-----------------------------------------------------------------------------
 * Config macros. Use these to config the driver.
 *---------------------------------------------------------------------------*/
#define MAX_FRAME_SIZE                 1532

#define CFG_RX_NUM_BUF_DESC            128
#define CFG_RX_NUM_BUF_SERVICE         8
#define CFG_RX_BUF_OFFSET              0
#define CFG_RX_BUF_MULTIPLIER          2

#define CFG_TX_NUM_BUF_DESC            128
#define CFG_TX_NUM_BUF_SERVICE         8 
#define CFG_TX_BUF_OFFSET              0     /* Lets not change this. */

/*-----------------------------------------------------------------------------
 * Private macros.
 *---------------------------------------------------------------------------*/
#define MAX_TIMER                      1
#define TICK_TIMER                     0

#define CHAN_CLOSE                     0

#define BLOCKING      		       1
#define CHAN_TEARDOWN                  2
#define CHAN_SETUP                     4
#define COMPLETE                       8
#define FREE_BUFFER                    16

#define MAX_TX_CHAN                    1
#define MAX_RX_CHAN                    7
#define PORT_BLOCK                     1
#define PORT_LEARN                     2
#define PORT_FWD                       3
#define ESWITCH_PORT_LINK             (1<<7)
#define ESWITCH_PORT_SPEED            (1<<6)
#define ESWITCH_PORT_DUPLEX           (1<<5)


typedef enum {
	       BRG_ERR = -1,
	       BRG_OK
             } BRG_STS;

typedef enum { ESWITCH_PHY_OP = 0,
	       ESWITCH_PORT_OP, 
	       ESWITCH_VLAN_INGRESS_OP, 
	       ESWITCH_DROP_UNTAG_OP,
	       ESWITCH_PORT_VLAN_OP,
	       ESWITCH_PORT_STP_OP,
               ESWITCH_MODE
               } ESWITCH_COMMAND; 

typedef enum {eswitch_auto_ngt = 1,
	      eswitch_hun_half,
	      eswitch_hun_full,
	      eswitch_ten_half,
	      eswitch_ten_full
             } PORT_CONFIG;

typedef enum {
	      e_Enable=0,
	      e_Disable
	     } ENBL_DSBL;	      

typedef enum {
	      ESWITCH_VLAN_ADD=1,
	      ESWITCH_VLAN_DEL	    
	     } ESWITCH_VLAN_COMMAND;	      

typedef enum {
	      ESWITCH_MULTICAST_ADD=1,
	      ESWITCH_MULTICAST_DEL
	     } ESWITCH_MCAST_COMMAND;	      

typedef struct SWITCHCONSCMD
{
	char name[20];
	union
	{
		struct VLANCMD
		{
			unsigned int  vid;
			unsigned char egr_ports;
			unsigned char untag_ports;
		}VlanCmd;        
		struct PORTCMD
		{
			unsigned int value;
			unsigned char port;
		}PortCmd;
		struct PRIOCMD
		{
			unsigned char port;
			unsigned char prio_in;
			unsigned char prio_out;
		}PrioCmd;		
	}CmdType;	

	unsigned char cmd;
	unsigned int  result;
} SwitchConsCmd;

typedef int  (*ESWITCH_HAL_ISR_FUNC_T)(HAL_DEVICE*, int*);

typedef struct
{
    void                    *owner;
    ESWITCH_HAL_ISR_FUNC_T  hal_isr;
    struct tasklet_struct   tasklet;
} ESWITCH_ISR_INFO_T;

typedef struct eswitch_chan
{
   int num_BD;
   int buffer_size;
   int buffer_offset;
   int service_max;
   int state;
   int tot_buf_size;
   int tot_reserve_bytes;
} ESWITCH_CHAN_T;

typedef struct 
{
    int             cfg_chan;
    int             dev_chan;
    int             opened_chan;
    ESWITCH_CHAN_T  chan[MAX_RX_CHAN];
} ESWITCH_RX_CHAN_INFO_T;

typedef struct
{
    int             cfg_chan;
    int             dev_chan;
    int             opened_chan;
    ESWITCH_CHAN_T  chan[MAX_TX_CHAN];
} ESWITCH_TX_CHAN_INFO_T;

typedef struct
{
   void                        *owner;
   HAL_FUNCTIONS               *hal_funcs;
   HAL_DEVICE                  *hal_dev;
   OS_FUNCTIONS                *os_funcs; 
} ESWITCH_DRV_HAL_INFO_T;


typedef struct
{
    unsigned long tx_discards;
    unsigned long rx_discards;
    unsigned long start_tick;
} ESWITCH_DRV_STATS_T;

typedef struct ESWITCH_PRIVATE_INFO_tag
{
    void                         *owner;
    struct timer_list            timer[1];
    ESWITCH_DRV_HAL_INFO_T       *drv_hal;
    ESWITCH_ISR_INFO_T           *eswitch_isr;
    ESWITCH_RX_CHAN_INFO_T       *rx_chan_info;
    ESWITCH_TX_CHAN_INFO_T       *tx_chan_info;
    ESWITCH_DRV_STATS_T          *stats;
    unsigned int 		 delay_ticks;
    char                         mac_addr[6];
    struct net_device_stats      net_dev_stats; 
    int                          set_to_close;
    unsigned long                dev_size;
    ESWITCH_CONFIG               *p_config;

} ESWITCH_PRIVATE_INFO_T;


void *eswitch_hal_malloc_buffer(unsigned int size, void *MemBase, unsigned int MemRange,
                                HAL_DEVICE *HalDev, HAL_RECEIVEINFO *HalReceiveInfo, 
                                OS_RECEIVEINFO **OsReceiveInfo, OS_DEVICE *OsDev);
void eswitch_hal_tear_down_complete(OS_DEVICE*, int, int);

int  eswitch_hal_control(OS_DEVICE *p_osdev, const char *key, 
                         const char *action, void *value);
int  eswitch_hal_receive(OS_DEVICE *p_osdev, FRAGLIST *fragList, 
                         unsigned int FragCount, unsigned int pkt_len, 
                         HAL_RECEIVEINFO *halReceiveInfo, 
                         unsigned int mode);

void  eswitch_hal_isr(int irq, void *p_param, struct pt_regs *p_cb_param);
void  eswitch_handle_tasklet(unsigned long data);

inline static int eswitch_ci_strcmp(const char *s1, const char *s2)
{
    while(*s1 && *s2)
    {
        if(tolower(*s1) != tolower(*s2))
            break;
        s1++;
        s2++;
    }
    return(tolower(*s1) - tolower(*s2));
}
#endif
