#define DIRECTION_TX 0
#define DIRECTION_RX 1

#define SETBIN 0
#define GETBIN 1
#define SETSTR 2
#define GETSTR 3

/* Channel Teardown Defines */

#define RX_TEARDOWN 2
#define TX_TEARDOWN 1
#define BLOCKING_TEARDOWN 8
#define FULL_TEARDOWN 4
#define PARTIAL_TEARDOWN 0

/*  HAL Data Structure definitions  */

#ifndef _CPSW_HAL
  typedef void HAL_DEVICE;
  typedef void HAL_RECEIVEINFO;
#endif

#include "cpswhalstddef.h"

typedef struct _hal_functions_ext_s HAL_FUNCTIONS_EXT;

/**
 *  @ingroup shared_data
 *  This is the fragment list structure.  Each fragment list entry contains a 
 *  length and a data buffer.
 */
typedef struct /*cpswhal.h                                                   */
  {
   bit32u   len;    /**< Length of the fragment in bytes (lower 16 bits are valid).  For SOP, upper 16 bits is the buffer offset. */
   void     *data;  /**< Pointer to fragment data */
   void     *OsInfo; /**< Pointer to OS defined data */
  }FRAGLIST;

#define Control(p1,p2,p3,p4) Control((p1),(p2),(p3),((void*)(p4)))

/**
 *  @ingroup shared_data
 *  The OS_FUNCTIONS struct defines the function pointers for all upper layer 
 *  functions accessible to the CPHAL.  The upper layer software is responsible 
 *  for providing the correct OS-specific implementations for the following 
 *  functions. It is populated by calling InitModule() (done by the CPHAL in 
 *  xxxInitModule().
 */
typedef struct /*cpswhal.h                                                   */
  {
  int   (*Control)(OS_DEVICE *OsDev, const char *Key, const char *Action, void *Value);
  void  (*CriticalOn)(void);
  void  (*CriticalOff)(void);
  void  (*DataCacheHitInvalidate)(void *MemPtr, int Size);
  void  (*DataCacheHitWriteback)(void *MemPtr, int Size);
  int   (*DeviceFindInfo)(int Inst, const char *DeviceName, void *DeviceInfo);
  int   (*DeviceFindParmUint)(void *DeviceInfo, const char *Parm, bit32u *Value);
  int   (*DeviceFindParmValue)(void *DeviceInfo, const char *Parm, void *Value);
  void  (*Free)(void *MemPtr);
  void  (*FreeRxBuffer)(OS_RECEIVEINFO *OsReceiveInfo, void *MemPtr);
  void  (*FreeDev)(void *MemPtr);
  void  (*FreeDmaXfer)(void *MemPtr);
  void  (*IsrRegister)(OS_DEVICE *OsDev, int (*halISR)(HAL_DEVICE*, int*), int InterruptBit);
  void  (*IsrUnRegister)(OS_DEVICE *OsDev, int InterruptBit);
  void* (*Malloc)(bit32u size);
  void* (*MallocDev)(bit32u Size);
  void* (*MallocDmaXfer)(bit32u size, void *MemBase, bit32u MemRange);
  void* (*MallocRxBuffer)(bit32u size, void *MemBase, bit32u MemRange,
                       OS_SETUP *OsSetup, HAL_RECEIVEINFO *HalReceiveInfo, 
                       OS_RECEIVEINFO **OsReceiveInfo, OS_DEVICE *OsDev);
  void* (*Memset)(void *Dest, int C, bit32u N);
  int   (*Printf)(const char *Format, ...);
  int   (*Receive)(OS_DEVICE *OsDev,FRAGLIST *FragList,bit32u FragCount,
                 bit32u PacketSize,HAL_RECEIVEINFO *HalReceiveInfo, bit32u Mode);
  int   (*SendComplete)(OS_SENDINFO *OsSendInfo);
  int   (*Sprintf)(char *S, const char *Format, ...);
  int   (*Strcmpi)(const char *Str1, const char *Str2);
  unsigned int (*Strlen)(const char *S);
  char* (*Strstr)(const char *S1, const char *S2);
  unsigned long  (*Strtoul)(const char *Str, char **Endptr, int Base);
  void  (*TeardownComplete)(OS_DEVICE *OsDev, int Ch, int Direction);
  } OS_FUNCTIONS;


/**
 *  @ingroup shared_data
 *  Channel specific configuration information.  This structure should be
 *  populated by upper-layer software prior to calling @c ChannelSetup().  Any
 *  configuration item that can be changed on a per channel basis should
 *  be represented here.  Each module may define this structure with additional
 *  module-specific members.  
 */
typedef struct 
  {
   int Channel;        /**< Channel number. */
   int Direction;      /**< Rx or Tx. */

   int RxBufSize;      /**< Size (in bytes) for each Rx buffer.*/
   int RxBufferOffset; /**< Number of bytes to offset rx data from start of buffer (must be less than buffer size). */
   int RxNumBuffers;   /**< The number of Rx buffer descriptors to allocate for Ch. */
/*   int RxServiceMax;   *< Maximum number of packets to service at one time. */

   int TxNumBuffers;   /**< The number of Tx buffer descriptors to allocate for Ch. */
   int TxNumQueues;    /**< Number of Tx queues for this channel (1-2). Choosing 2 enables a low priority SAR queue. */
/*   int TxServiceMax;   *< Maximum number of packets to service at one time. */   

  int   (*Receive)(OS_DEVICE *OsDev,FRAGLIST *FragList,bit32u FragCount,
                 bit32u PacketSize,HAL_RECEIVEINFO *HalReceiveInfo, bit32u Mode);
  int   (*SendComplete)(OS_SENDINFO *OsSendInfo);                    /*M001*/

/*Must add channel info extension...                                         */
  }CHANNEL_INFO;

/**
 *  @ingroup shared_data
 *  The HAL_FUNCTIONS struct defines the function pointers used by upper layer 
 *  software.  The upper layer software receives these pointers through the
 *  call to xxxInitModule().
 */
typedef struct /*cpswhal.h                                                   */
  {
  int  (*ChannelSetup)    (HAL_DEVICE *HalDev, CHANNEL_INFO *Channel, OS_SETUP *OsSetup);
  int  (*ChannelTeardown) (HAL_DEVICE *HalDev, int Channel, int Mode);
  int  (*Close)           (HAL_DEVICE *HalDev, int Mode);
  int  (*Control)         (HAL_DEVICE *HalDev, const char *Key, int Action, void *Value);
  int  (*Init)            (HAL_DEVICE *HalDev);
  int  (*Open)            (HAL_DEVICE *HalDev);
  int  (*PacketProcessEnd) (HAL_DEVICE *HalDev);
  int  (*Probe)           (HAL_DEVICE *HalDev);
  int  (*RxReturn)        (HAL_RECEIVEINFO *HalReceiveInfo, int StripFlag);
  int  (*Send)            (HAL_DEVICE *HalDev,FRAGLIST *FragList,int FragCount,int PacketSize, OS_SENDINFO *OsSendInfo,int (*SendComplete)(OS_SENDINFO *),bit32u Mode);
  int  (*Shutdown)        (HAL_DEVICE *HalDev);
  int  (*Tick)            (HAL_DEVICE *HalDev);
  HAL_FUNCTIONS_EXT       *HalFuncExt;
#ifdef _CPHAL_AAL5
  int  (*Kick) (HAL_DEVICE *HalDev, int Queue);
  volatile bit32u* (*RegAccess)(HAL_DEVICE *HalDev, bit32u RegOffset);
  STAT_INFO*  (*StatsGetOld)(HAL_DEVICE *HalDev);
#endif
  } HAL_FUNCTIONS;

#define cpswhalPushBin(key,val) HalFunc->Control(HalDev,key,SETBIN,val)

extern const char hcCpuFrequency[];
extern const char hcResetBase[];
extern const char hcdebug[];
extern const char hcdev_base[];
extern const char hcMaxChannel[];
extern const char hcTxServiceMax[];
extern const char hcRxServiceMax[];
extern const char hcid[];
extern const char hcoffset[];
extern const char hcInst[];
extern const char hcTick[];
extern const char hcSet[];
extern const char hcClear[];
extern const char hcStateChange[];
