/**@file*********************************************************************
 *  TNETDxxxx Software Support
 *  Copyright (c) 2003 Texas Instruments Incorporated. All Rights Reserved.
 *
 *  FILE: cpswhal.h
 *
 *  DESCRIPTION:
 *      User include file, contains data definitions shared between the CPHAL
 *      and the upper-layer software.
 *
 *  HISTORY:
 *      Date      Modifier  Notes
 *      06May03   Mick      From CPHAL 1.5
 *      14Nov03   Mick      stddef replaced with defs.h
 *****************************************************************************/

#ifndef _CPSWHAL_H
#define _CPSWHAL_H

#include "cpswhalcommon_defs.h"
#include "cpswhalcommon_ecerrors.h"
#include "cpswhalcommon_global.h"

/*  HAL Data Structure definitions for CPHAL users */

#ifndef _CPSW_HAL
  typedef void HAL_DEVICE;
  typedef void HAL_RECEIVEINFO;
#endif

/*
 *
 *  MDIO definitions
 *
*/

#define SNWAY_AUTOMDIX       (1<<16)   /* Bit 16 and above not used by MII register */
#define SNWAY_FD1000         (1<<13)
#define SNWAY_HD1000         (1<<12)
#define SNWAY_NOPHY          (1<<10)
#define SNWAY_LPBK           (1<<9)
#define SNWAY_FD100          (1<<8)
#define SNWAY_HD100          (1<<7)
#define SNWAY_FD10           (1<<6)
#define SNWAY_HD10           (1<<5)
#define SNWAY_AUTO           (1<<0)
#define SNWAY_AUTOALL        (SNWAY_AUTO|SNWAY_FD100|SNWAY_FD10|SNWAY_HD100|SNWAY_HD10)

#define DIRECTION_TX 0
#define DIRECTION_RX 1

/* Channel Teardown Defines */

#define PARTIAL_TEARDOWN  0
#define TX_TEARDOWN       1
#define RX_TEARDOWN       (1<<1)
#define FULL_TEARDOWN     (1<<2)
#define BLOCKING_TEARDOWN (1<<3)

/**
 *  @ingroup shared_data
 *  These are the actions defined for use with the @c Control() HAL function.
 *  Not all actions are applicable to all @p HAL_CONTROL_KEYs.  Refer to the
 *  module specific documentation in the appendices to determine which actions
 *  may be used to access a given @p HAL_CONTROL_KEY.
 */
enum HAL_CONTROL_ACTION
  {
   haCLEAR=0,      /**< Clears the value of the key. (See specific Key for allowable use) */
   haGETbyREF,     /**< Gets the value of the key by reference. */
   haGETDefault,   /**< Returns the Default for the key by reference */
   haGETMax,       /**< Returns the Maximum Range for the key by reference */
   haGETMin,       /**< Returns the Minimum Range for the key by reference */
   haNULL,         /**< No action defined (used when a key does not represent data, but is used as a signal). */
   haSETbyREF,     /**< Sets the value of the key by reference. */
   haSETbyVAL,     /**< Sets the value of the key by value. */
   haSETbySTR      /**< @internal */
  };
typedef enum HAL_CONTROL_ACTION HAL_CONTROL_ACTION;

/*
 *
 *   P U S H  /  P U L L  M A C R O S
 *
*/

#define halIoctl(action, key, value) HalFunc->Control(HalDev,key,action,value)
#define halChMacro(key,value,ch,ChMacro) \
{ \
   char *FullKey; \
   FullKey = (char *)OsDev->OsFunc->Malloc(40); \
   OsDev->OsFunc->Sprintf(FullKey, "%s.%d", key, ch); \
   ChMacro(FullKey, value); \
   OsDev->OsFunc->Free(FullKey); \
}
#define halChQMacro(key,value,ch,q,ChMacro) \
{ \
   char *FullKey; \
   FullKey = (char *)OsDev->OsFunc->Malloc(40); \
   OsDev->OsFunc->Sprintf(FullKey, "%s.%d.%d", key, ch, q); \
   ChMacro(FullKey, value); \
   OsDev->OsFunc->Free(FullKey); \
}

/**
@defgroup Parameter_Macros Useful Macros for Parameter Access

The macros defined here may be used by the OS to access HAL
parameters (HAL_CONTROL_KEYs) in a simple fashion.  NOTE: Each of these
macros requires certain local variables to be in scope.  All of the
macros require the following two variables: 'HAL_FUNCTIONS *HalFunc', and
'HAL_DEVICE *HalDev'.  The macros which accept a channel value require
the variable 'OS_DEVICE *OsDev' to be in scope.
@{
*/

/** @hideinitializer For HAL_CONTROL_KEY 'key', gets the 'value' by reference.  Caller must
    provide a pointer to the expected type in 'value'.  Equivalent to the haGETbyREF action. */
#define halPullRef(key,value)   halIoctl(haGETbyREF,   key, value)
/** @hideinitializer For HAL_CONTROL_KEY 'key', sets the 'value' by reference.  Caller must
    provide a pointer to the data in 'value'.  Equivalent to the haSETbyREF action. */
#define halPushRef(key,value)   halIoctl(haSETbyREF,   key, value)
/** @hideinitializer For HAL_CONTROL_KEY 'key', sets the 'value' by passing the actual value.
    Equivalent to the haSETbyVAL action. */
#define halPushValue(key,value) halIoctl(haSETbyVAL, key, (void*)(value))

/* The following macros allow the OS to easily set/get a Channel or Channel/Queue parameter value. */

/** @hideinitializer For HAL_CONTROL_KEY 'key' and (int) channel 'ch', sets the 'value' by passing the actual value.
    Equivalent to the haSETbyVAL action. */
#define halPushValueCh(key,value,ch) halChMacro(key,value,ch,halPushValue)

/** @hideinitializer For HAL_CONTROL_KEY 'key' and (int) channel 'ch', sets the 'value' by reference.  Caller must
    provide a pointer to the data in 'value'.  Equivalent to the haSETbyREF action. */
#define halPushRefCh(key,value,ch) halChMacro(key,value,ch,halPushRef)

/** @hideinitializer For HAL_CONTROL_KEY 'key' and (int) channel 'ch', gets the 'value' by reference.  Caller must
    provide a pointer to the expected type in 'value'.  Equivalent to the haGETbyREF action. */
#define halPullRefCh(key,value,ch) halChMacro(key,value,ch,halPullRef)

/** @hideinitializer For HAL_CONTROL_KEY 'key', (int) channel 'ch', and (int) queue 'q', sets the 'value' by passing the actual value.
    Equivalent to the haSETbyVAL action. */
#define halPushValueChQ(key,value,ch,q) halChQMacro(key,value,ch,q,halPushValue)

/** @hideinitializer For HAL_CONTROL_KEY 'key', (int) channel 'ch', and (int) queue 'q', sets the 'value' by reference.  Caller must
    provide a pointer to the data in 'value'.  Equivalent to the haSETbyREF action. */
#define halPushRefChQ(key,value,ch,q) halChQMacro(key,value,ch,q,halPushRef)

/** @hideinitializer For HAL_CONTROL_KEY 'key', (int) channel 'ch', and (int) queue 'q', gets the 'value' by reference.  Caller must
    provide a pointer to the expected type in 'value'.  Equivalent to the haGETbyREF action. */
#define halPullRefChQ(key,value,ch,q) halChQMacro(key,value,ch,q,halPullRef)

/** @hideinitializer For (int) level 'lvl', returns a group of statistics values.  Caller must provide
    an integer pointer as the 'value' parameter.  Please refer to Appendix A, Configuration and Control, for
    more information about statistics groups.  */
#define halGetStats(lvl,value) \
  { \
   char *FullKey; \
   FullKey = (char *)OsDev->OsFunc->Malloc(40); \
   OsDev->OsFunc->Sprintf(FullKey, "Stats.%d", lvl); \
   halPullRef(FullKey, value); \
   OsDev->OsFunc->Free(FullKey); \
  }

/** @hideinitializer For (int) level 'lvl', (int) channel 'ch', returns a group of statistics
    values.  Caller must provide an integer pointer as the 'value' parameter.  Please refer to Appendix A,
    Configuration and Control, for more information about statistics groups.  */
#define halGetStatsCh(lvl,value,ch) \
  { \
   char *FullKey; \
   FullKey = (char *)OsDev->OsFunc->Malloc(40); \
   OsDev->OsFunc->Sprintf(FullKey, "Stats.%d.%d", lvl, ch); \
   halPullRef(FullKey, value); \
   OsDev->OsFunc->Free(FullKey); \
  }

/** @hideinitializer For (int) level 'lvl', (int) channel 'ch', (int) queue 'q', returns a group of statistics
    values.  Caller must provide an integer pointer as the 'value' parameter.  Please refer to Appendix A,
    Configuration and Control, for more information about statistics groups.  */
#define halGetStatsChQ(lvl,value,ch,q) \
  { \
   char *FullKey; \
   FullKey = (char *)OsDev->OsFunc->Malloc(40); \
   OsDev->OsFunc->Sprintf(FullKey, "Stats.%d.%d.%d", lvl, ch, q); \
   halPullRef(FullKey, value); \
   OsDev->OsFunc->Free(FullKey); \
  }

/** @} */

#define osIoctl(action, key, value) OsFunc->Control(OsDev,key,action,value)
#define osPullRef(key,value)        osIoctl(haGETbyREF, key, value)
#define osPushRef(key,value)        osIoctl(haSETbyREF, key, value)
#define osPushValue(key,value)      osIoctl(haSETbyVAL, key, (void*)(value))

typedef struct hal_functions_ext_s  HAL_FUNCTIONS_EXT;
typedef struct channel_info_ext_s   CHANNEL_INFO_EXT;
typedef struct os_functions_ext_s   OS_FUNCTIONS_EXT;

/**
@defgroup shared_data Shared Data Structures

The data structures documented here are shared by all modules.
*/

/**
 *  @ingroup shared_data
 *  This is the fragment list structure.  Each fragment list entry contains a
 *  length and a data buffer.
 */
typedef struct
{
   bit32u   len;    /**< Length of the fragment in bytes (lower 16 bits are valid).  For SOP, upper 16 bits is the buffer offset. */
   void     *data;  /**< Pointer to fragment data */
   void     *OsInfo; /**< Pointer to OS defined data */
}FRAGLIST;


/**
 *  @ingroup shared_data
 *  Channel specific configuration information.  This structure should be
 *  populated by upper-layer software prior to calling @c ChannelSetup().  Any
 *  configuration item that can be changed on a per channel basis should
 *  be represented here.  Each module may define this structure with additional
 *  module-specific members through the Ext pointer.
 */
typedef struct
{
   int Channel;        /**< Channel number. */
   int Direction;      /**< Use the supplied #define DIRECTION_RX or DIRECTION_TX. */
   /**
    *  Per channel @c osReceive() function pointer.  See the @link OS_Functions OS Functions API
    *  @endlink documentation for complete information.
    */
   int   (*Receive)(OS_DEVICE *OsDev,FRAGLIST *FragList,bit32u FragCount,
                 bit32u PacketSize,HAL_RECEIVEINFO *HalReceiveInfo, bit32u Mode);
    /** May be used by each module to add additional channel configuration items.  If used,
        the documentation for such items will be located in the module specific appendices. */
   CHANNEL_INFO_EXT *Ext;
}CHANNEL_INFO;



/**
 *  @ingroup shared_data
 *  The HAL_FUNCTIONS struct defines the function pointers used by upper layer
 *  software.  The upper layer software receives these pointers through the
 *  call to cpswHal<module>InitModule().
 */
typedef struct
{
  int  (*ChannelSetup)     (HAL_DEVICE *HalDev, CHANNEL_INFO *Channel, OS_SETUP *OsSetup);
  int  (*ChannelTeardown)  (HAL_DEVICE *HalDev, int Channel, int Mode);
  int  (*Close)            (HAL_DEVICE *HalDev, int Mode);
  int  (*Control)          (HAL_DEVICE *HalDev, HAL_CONTROL_KEY *Key, HAL_CONTROL_ACTION Action, void *Value);
  int  (*Init)             (HAL_DEVICE *HalDev);
  int  (*Open)             (HAL_DEVICE *HalDev);
  int  (*PacketProcessEnd) (HAL_DEVICE *HalDev);
  int  (*Probe)            (HAL_DEVICE *HalDev);
  int  (*RxReturn)         (HAL_RECEIVEINFO *HalReceiveInfo, int StripFlag);
  int  (*Send)             (HAL_DEVICE *HalDev,FRAGLIST *FragList,int FragCount,int PacketSize, OS_SENDINFO *OsSendInfo,int (*SendComplete)(OS_SENDINFO *),bit32u Mode);
  int  (*Shutdown)         (HAL_DEVICE *HalDev);
  int  (*Tick)             (HAL_DEVICE *HalDev);
  /** Each module may define extra HAL functions through the Ext pointer.  See the appropriate module specific
      appendix for related documentation. */
  HAL_FUNCTIONS_EXT        *Ext;
} HAL_FUNCTIONS;


/**
 *  @ingroup shared_data
 *  The OS_FUNCTIONS struct defines the function pointers for all upper layer
 *  functions accessible to the CPHAL.  The upper layer software is responsible
 *  for providing the correct OS-specific implementations for the following
 *  functions. It is populated by calling InitModule() (done by the CPHAL in
 *  cpswHal<module>InitModule().
 */
typedef struct
{
  int   (*Control)                    (OS_DEVICE *OsDev, HAL_CONTROL_KEY *Key, HAL_CONTROL_ACTION Action, void *Value);
  void  (*CriticalOn)                 (void);
  void  (*CriticalOff)                (void);
  void  (*DataCacheHitInvalidate)     (void *MemPtr, int Size);
  void  (*DataCacheHitWriteback)      (void *MemPtr, int Size);
  int   (*DeviceFindInfo)             (int Inst, const char *DeviceName, void *DeviceInfo);
  int   (*DeviceFindParmUint)         (void *DeviceInfo, const char *Parm, bit32u *Value);
  int   (*DeviceFindParmValue)        (void *DeviceInfo, const char *Parm, void *Value);
  void  (*DeviceFindInfoRelease)      (void *DeviceInfo);
  void  (*DeviceFindParmValueRelease) (void *Value);
  void  (*Free)                       (void *MemPtr);
  void  (*FreeRxBuffer)               (OS_RECEIVEINFO *OsReceiveInfo, void *MemPtr);
  void  (*FreeDev)                    (void *MemPtr);
  void  (*FreeDmaXfer)                (void *MemPtr);
  void  (*IsrRegister)                (OS_DEVICE *OsDev, int (*halISR)(HAL_DEVICE*, int*), int InterruptBit);
  void  (*IsrUnRegister)              (OS_DEVICE *OsDev, int InterruptBit);
  void* (*Malloc)                     (bit32u size);
  void* (*MallocDev)                  (bit32u Size);
  void* (*MallocDmaXfer)              (bit32u size, void *MemBase, bit32u MemRange);
  void* (*MallocRxBuffer)             (bit32u size, void *MemBase, bit32u MemRange,
                                       OS_SETUP *OsSetup, HAL_RECEIVEINFO *HalReceiveInfo,
                                       OS_RECEIVEINFO **OsReceiveInfo, OS_DEVICE *OsDev);
  int   (*Printf)                     (const char *Format, ...);
  int   (*Sprintf)                    (char *S, const char *Format, ...);
  void  (*TeardownComplete)           (OS_DEVICE *OsDev, int Ch, int Direction);
  /** The OS may define extra OS functions through the Ext pointer. */
  OS_FUNCTIONS_EXT *Ext;
} OS_FUNCTIONS;

#endif
