/******************************************************************************
 *  TNETDxxxx Software Support
 *  Copyright (c) 2003 Texas Instruments Incorporated. All Rights Reserved.
 *
 *  FILE:
 *
 *  DESCRIPTION:
 *   This file contains all the global constructs for the CPSW HAL
 *
 *  HISTORY:
 *   Date      Modifier      Ver    Notes
 *  01Oct03   Greg/Denis   1.00    Original Version
 *****************************************************************************/

/* Needed for the cpswhal.h include being done */

typedef void OS_DEVICE;
typedef void OS_SENDINFO;
typedef void OS_RECEIVEINFO;
typedef void OS_SETUP;

#include "cpswhalcommon_params.h"
#include "cpswhalcommon_support.h"

/*
 *  Checks a parameter in the parameter table to see if we are allowed
 *  to write it, given the current state of the system and the
 *  permissions related to the parameter.  Returns TRUE if writable,
 *  FALSE if not.
 */

static int IsWritable(HAL_DEVICE *HalDev, const DEVICE_PARAMS *DeviceParams, SUB_STRUCT *SubStruct, int StructIndex, int Ch)
{
    /* determine if permissible to write */
    if(!(DeviceParams->Type & WRITABLE))
      return(FALSE);   /* param not writable */

    /*  Parameter is writable, if can write after open, then done */
    if(DeviceParams->Type & WR_AFTER_OPEN)
      return(TRUE);    /* writable, even with Module Open*/

    /* Parameter is writable but not after open, check if module Open */
    if(SubStruct[StructIndex].IsOpen(HalDev, Ch, CH_DIRECTION(DeviceParams->Type)))
      return (FALSE);  /* Module is open, cannot update parameter */
    else
      return(TRUE);    /* Module is not open, can update parameter */
}

/*
 *  Determines the reset register address to be used for a particular device.
 *  It will search the current device entry for Locator information.  If the
 *  device is a root device, there will be no Locator information, and the
 *  function will find and return the root reset register.  If a Locator value
 *  is found, the function will search each VLYNQ device entry in the system
 *  looking for a matching Locator.  Once it finds a VLYNQ device entry with
 *  a matching Locator, it will extract the "ResetBase" parameter from that
 *  VLYNQ device entry (thus every VLYNQ entry must have the ResetBase parameter).
 *
 *  @param  HalDev   CPHAL module instance. (set by cpswHal<module>InitModule())
 *  @param  ResetBase Pointer to integer address of reset register.
 *
 *  @return 0 OK, Non-zero not OK
 */
static int ResetBaseGet(OS_FUNCTIONS *OsFunc, bit32u *ResetBase, char *DeviceInfo)
  {
   char *VlynqDeviceInfo;
   char *DeviceLocator, *VlynqLocator;
   int Inst=0;
   bit32u error_code;

   error_code = OsFunc->DeviceFindParmValue(DeviceInfo, hkLocator, &DeviceLocator);
   if (error_code)
     {
      /* if no Locator value, device is on the root, so get the "reset" device */
      error_code = OsFunc->DeviceFindInfo(0, hkReset, &DeviceInfo);
      if  (error_code)
        {
         return(EC_VAL_DEVICE_NOT_FOUND);
        }

      error_code = OsFunc->DeviceFindParmUint(DeviceInfo, hkBase, ResetBase);
      if (error_code)
        {
         return(EC_VAL_BASE_ADDR_NOT_FOUND);
        }

      *ResetBase = ((bit32u)PhysToVirtNoCache(*ResetBase));

      /* release DeviceInfo */
      OsFunc->DeviceFindInfoRelease(DeviceInfo);

      /* found base address for root device, so we're done */
      return (EC_NO_ERRORS);
     }
    else
     {
      /* we have a Locator value, so the device is remote */

      /* Find a vlynq device with a matching locator value */
      while ((OsFunc->DeviceFindInfo(Inst++, hkVlynq, &VlynqDeviceInfo)) == EC_NO_ERRORS)
        {
         error_code = OsFunc->DeviceFindParmValue(VlynqDeviceInfo, hkLocator, &VlynqLocator);
         if (error_code)
           {
            /* no Locator value for this VLYNQ, so move on */
            continue;
           }

         if (cpswHalCommonstrcmpi(DeviceLocator, VlynqLocator)==0)
           {
            /* we have found a VLYNQ with a matching Locator, so extract the ResetBase */
            error_code = OsFunc->DeviceFindParmUint(VlynqDeviceInfo, hkResetBase, ResetBase);
            if (error_code)
              {
               return(EC_VAL_BASE_ADDR_NOT_FOUND);
              }
            *ResetBase = ((bit32u)PhysToVirtNoCache(*ResetBase));

            /* found base address for root device, so we're done */
            return (EC_NO_ERRORS);
           }

         /* release DeviceInfo, VlynqLocator */
         OsFunc->DeviceFindInfoRelease(VlynqDeviceInfo);
         OsFunc->DeviceFindParmValueRelease(VlynqLocator);
        } /* while */
     } /* else */

   /* release DeviceLocator */
   OsFunc->DeviceFindParmValueRelease(DeviceLocator);

   return (EC_NO_ERRORS);
  }

/**
 *  @ingroup Params_Functions
 *
 *  Returns statistics information for a particular group of stats (selected by @p Index).
 *  Information is returned via StatPtr.  This function allocates a block of memory in the following
 *  format (assuming 2 statistics returned):
 *
 *  Byte addr 0 -> Pointer to Stat 1 Name (char *)
 *            4 -> Pointer to Stat 1 Value
 *            8 -> Pointer to Stat 2 Name (char *)
 *           12 -> Pointer to Stat 2 Value
 *           16 -> Empty
 *           20 -> String representing Stat 1 Value (pointed to by address 4)
 *           20 + Prev string len + 1 (null ch) -> String for Stat 2 Value
 *
 *  All values are represented in string form as shown above.  The caller is responsible
 *  for freeing the block of memory that this function allocates at *StatPtr.
 *
 *  @param  OsFunc   Os function block pointer.
 *  @param  DeviceParams  Pointer to the device params table.
 *  @param  SubStruct     Pointer to substructure definitions for this module.
 *  @param  StatPtr       The output parameter, passed by reference.
 *  @param  Index         Statistics group to collect.
 *  @param  Ch            Channel number (for statistics groups indexed by channel)
 *  @param  Queue         Queue number (for statistics groups indexed by queue)
 *
 *  @return EC_NO_ERRORS
 */
int cpswHalCommonStatsGet(OS_FUNCTIONS *OsFunc, const DEVICE_PARAMS *DeviceParams, SUB_STRUCT *SubStruct,void **StatPtr, int Index, int Ch, int Queue)
  {
   int Size;
   bit32u CurrentAddr;
   bit32u *AddrPtr;
   char *DataPtr;
   int NumberOfStats=0;
   const DEVICE_PARAMS *ParmTable = DeviceParams;
   void *stbase;
   int StructIndex;

   /* determine number of parameters */
   while(DeviceParams->pszKey)
     {
      if (((DeviceParams->Type&GROUP_MASK)>>GROUP_SHIFT) == Index)
        {
         NumberOfStats++;
        }
      DeviceParams++;
     }

   /*
      Reserve space for 3 items per stat (one name pointer, one data pointer,
      and one string representation of the value).  The value is stored in string
      form which can be up to 11 bytes long, in the form "0xabcdef01\0"
   */
   Size = sizeof(bit32u)*((NumberOfStats*2)+1);
   Size += (NumberOfStats*11); /* 11 bytes is to hold a string like "0xabcdef01\0" */
   *StatPtr = (bit32u *)OsFunc->Malloc(Size);

   AddrPtr = (bit32u *) *StatPtr;
   DataPtr = (char *)AddrPtr;
   DataPtr += sizeof(bit32u)*((NumberOfStats*2)+1);

   /* Loop through the parameter table looking for parameters that are in the requested
      statistics group.  */
   while(ParmTable->pszKey)
     {
      if (((ParmTable->Type&GROUP_MASK)>>GROUP_SHIFT) == Index)
        {
         StructIndex = (ParmTable->Type&SUBSTRUCT_MASK)>>SUBSTRUCT_SHIFT;
         stbase = SubStruct[StructIndex].sub_struct;

         CurrentAddr = (((bit32u)stbase)+((bit32u)ParmTable->Offset));

         *AddrPtr++ = (bit32u)ParmTable->pszKey;
         *AddrPtr++ = (bit32u)DataPtr;
         
         if ((ParmTable->Type&ISSTRING)!=0)
           {
            OsFunc->Sprintf(DataPtr, "%s", *((char **)(CurrentAddr +
              (SubStruct[StructIndex].sub_struct_size*(Ch*SubStruct[StructIndex].sub_index_size2)+Queue))));
           }
          else
           {
            OsFunc->Sprintf(DataPtr, "0x%x", *((bit32u *)(CurrentAddr +
              (SubStruct[StructIndex].sub_struct_size*(Ch*SubStruct[StructIndex].sub_index_size2)+Queue))));
           }
    
         DataPtr += cpswHalCommonstrlen(DataPtr) + 1;
        }
      ParmTable++;
     }

   *AddrPtr = (bit32u) 0;

   return (EC_NO_ERRORS);
  }

/* NOTE: Should this be in support rather than params? */
/* function looks for a decimal integer, maximum 19 digits in length */
int cpswHalCommonParseValue(const char *invalue, bit32u *value, int index)
  {
   char *p;
   char ValueFound[20];

   p = &ValueFound[0];

   while(*invalue)
     {
       if(*invalue != ';')
         {
         *p++ = *invalue++;
         }
       else
         {
           index--;
           if(index==0)
             {
               break;  /* Value Found */
             }
           else
             {
              p = &ValueFound[0]; /* not value wanted, reset pointer */
              invalue++;
             }
         }
     } /* while loop */

   if(index>1)
     return(1);  /* search terminated before delimiter found for requested index*/

   *p = '\0';  /* null terminate found value */

   *value = cpswHalCommonstrtoul(&ValueFound[0], 0, 10); /* convert value */
   return(0);
  }

void cpswHalCommonParseKey(const char *inkey,
                       char **str1,
                       char **str2,
                       char **str3,
                       int *index1,
                       int *index2)
  {
   *str1=(char *)inkey;
   *str3=0;
   *str2=cpswHalCommonstrchr(*str1,'.');
   if (*str2)
     {
      *(*str2)++=0;
      *index1 = cpswHalCommonstrtoul(*str2,0,10);
      *str3=cpswHalCommonstrchr(*str2,'.');
      if (*str3)
        {
         *(*str3)++=0;
         *index2 = cpswHalCommonstrtoul(*str3,0,10);
        }
     }
  }

/**
 *  @ingroup Params_Functions
 *

 *  @param  HalDev        Device handle.
 *  @param  OsFunc        Os function block pointer.
 *  @param  DeviceParams  Pointer to the device params table.
 *  @param  SubStruct     Pointer to substructure definitions for this module.
 *  @param  Key           Pointer to Key string, representing which parameter to access.
 *  @param  Action        Indicates the action to perform on the parameter (haGETbyREF, haSETbyREF, haSETbyVAL)
 *  @param  Value         (Void *) Output parameter for GET actions, input parameter for SET actions.
 *
 *
 *  @return EC_NO_ERRORS
 *
 */
int cpswHalCommonParamsControl(HAL_DEVICE *HalDev,OS_FUNCTIONS *OsFunc,const DEVICE_PARAMS *DeviceParams,SUB_STRUCT *SubStruct, const char *Key, int Action, void *Value)
  {
  char *p1,*p2,*p3;
  void *stbase;
  int StructIndex, Ch=0, Queue=0,loops,i;
  int sub_struct_size, sub_index_size1, sub_index_size2;
  int rc;

  /* Default retrun code to not finding the Key */

  rc = EC_VAL_KEY_NOT_FOUND;

#ifdef __CPHAL_DEBUG
  if (Key == hkDumpParams)
    {
     cpswHalCommonParamsDump(HalDev,OsFunc,DeviceParams,SubStruct);
     return(0);
    }
#endif

  if(Key == hkVersionCommon)
  {
      if(Action==haGETbyREF)
      {
          *((char **)Value) = (char *)hkVersion_COMMON;
          return(0);
      }
      else
      {
          return(EC_FUNC_CONTROL|EC_VAL_INVALID_ACTION);
      }
  }

  cpswHalCommonParseKey(Key, &p1, &p2, &p3, &Ch, &Queue);

  /*  Loop to find Match */
  while(DeviceParams->pszKey)
    {
    if ((cpswHalCommonstrcmp(DeviceParams->pszKey,p1)==0))
      {
       /* Match Found, if no variable assigned, break out of Find Match loop */
       if(DeviceParams->Type & PASS_THROUGH)
         break;
       StructIndex = (DeviceParams->Type&SUBSTRUCT_MASK)>>SUBSTRUCT_SHIFT;
       sub_struct_size = SubStruct[StructIndex].sub_struct_size;
       sub_index_size1 = SubStruct[StructIndex].sub_index_size1;
       sub_index_size2 = SubStruct[StructIndex].sub_index_size2;

       /* calculate address of first parameter to be accessed */
       stbase = SubStruct[StructIndex].sub_struct + ((bit32u)DeviceParams->Offset);
       ((bit32u)stbase) += (sub_struct_size * Ch * sub_index_size2) + (sub_struct_size * Queue);

       /* NOTE: All "Get" operations return a single value, Ch and Queue default to 0 */
       if(Action==haGETMax)
       {
          *(bit32u*)Value = DeviceParams->HighValue;
          rc = 0;
          break;
       }
       if(Action==haGETMin)
       {
          *(bit32u*)Value = DeviceParams->LowValue;
          rc = 0;
          break;
       }
       if(Action==haGETbyREF)
         {
         if(DeviceParams->Type & ISSTRING)
           {
           *((char **)Value)=*(char **)(((bit32u)stbase));
           }
         else
           {
           *((bit32u*)Value)=*((bit32u*)(((bit32u)stbase)));
           }
         rc = 0;
         break;
         }

       if (Action == haCLEAR)
         {
          rc = 0; break;
         }

       if ((Action==haSETbyVAL) || (Action==haSETbyREF) || (Action==haSETbySTR))
         {
          /* calculate the number of parameters to be accessed */
          if (p2)
            {
             if (p3)
               loops = 1; /* caller specified both indices */
              else
               loops = sub_index_size2; /* caller specified first index, so # is size of second index */
            }
           else
            {
             loops = sub_index_size1 * sub_index_size2; /* caller specified nothing, so set everything */
            }

          /* if this is a SET, we must find out the current state of the substruct so that we can determine, based on parm table rules, whether or not the SET is currently allowable */
          if (!(IsWritable(HalDev, DeviceParams, SubStruct, StructIndex, Ch)))
            {
             return(EC_FUNC_CONTROL|EC_VAL_PARAM_NOT_WRITABLE);
            }

          for (i=0; i<loops; i++)
            {
             if ((Action==haSETbyVAL)&&((DeviceParams->Type&ISSTRING)==0))
               {
                if (!( ( DeviceParams->LowValue<=((bit32u)Value) ) &&( DeviceParams->HighValue>=((bit32u)Value) ) ) )
                   return(EC_VAL_OUT_OF_RANGE);

                if ((DeviceParams->Type&ISBASE) != 0)
                  {
                   /* if this is a base address, assure that we convert to virtual, uncached */
                   *((bit32u*)(((bit32u)stbase)))=(bit32u)PhysToVirtNoCache((bit32u)Value);
                  }
                 else
                  {
                   *((bit32u*)(((bit32u)stbase)))=(bit32u)Value;
                  }
                rc = 0;
               }

             if ((Action==haSETbyREF)&&((DeviceParams->Type&ISSTRING)==0))
               {
                if(!( ( DeviceParams->LowValue<=(*(bit32u*)Value) ) &&( DeviceParams->HighValue>=(*(bit32u*)Value) ) ) )
                   return(EC_VAL_OUT_OF_RANGE);

                if((DeviceParams->Type&ISBASE) != 0)
                  {
                   /* if this is a base address, assure that we convert to virtual, uncached */
                   *((bit32u*)(((bit32u)stbase)))=(bit32u)PhysToVirtNoCache(*(bit32u*)Value);
                  }
                 else
                  {
                   *((bit32u*)(((bit32u)stbase)))=*(bit32u*)Value;
                  }
                rc = 0;
               } /* haSETP logic */

             if ((Action==haSETbyVAL)&&((DeviceParams->Type&ISSTRING)!=0))
               {
                /* SETbyVAL for strings uses a string copy */
                cpswHalCommonstrcpy( (char *)(((bit32u)stbase)), Value);
                rc = 0;
               }

             if ((Action==haSETbyREF)&&((DeviceParams->Type&ISSTRING)!=0))
               {
                /* SETbyREF for strings uses a pointer copy */
                *((char**)((bit32u)stbase))=(char *)Value;
                rc = 0;
               }

             if (Action==haSETbySTR)
               {
                if((DeviceParams->Type&ISSTRING)!=0)
                  {
                   *((char**)((bit32u)stbase))=(char *)Value;
                  }
                 else
                  {
                   if ((DeviceParams->Type&ISBASE) != 0)
                     {
                      /* if this is a base address, assure that we convert to virtual, uncached */
                      *((bit32u*)(((bit32u)stbase)))=(bit32u)PhysToVirtNoCache((bit32u)cpswHalCommonatoul(Value));
                     }
                    else
                     {
                      *((bit32u*)(((bit32u)stbase)))=(bit32u)cpswHalCommonatoul(Value);
                     }
                  }
                rc = 0;
               }
             /* sub_struct_size must be defaulted to one so that integer arrays within HalDev and other
                non-indexed structures can be correctly accessed */

             stbase += sub_struct_size;
            } /* for */
          /* break out after the for loop */
          break;
         } /* if haSETbyVAL || haSETSTR */

       return(EC_VAL_INVALID_ACTION);
      } /* if key matches (or pass through) */
     DeviceParams++;
    } /* while */

   /* call special function if provided */
   if (DeviceParams->ParamFunc != 0)
     {
      rc = DeviceParams->ParamFunc(HalDev, Key, Action, Value);
     }

   return(rc);
  }

#ifdef __CPHAL_DEBUG
int cpswHalCommonParamsDump(HAL_DEVICE *HalDev,OS_FUNCTIONS *OsFunc,const DEVICE_PARAMS *DeviceParams, SUB_STRUCT *SubStruct)
  {
   int line=0, StructIndex,sub_struct_size,sub_index_size1,sub_index_size2,i,j;
   void *stbase;

   while(DeviceParams->pszKey)
     {
      StructIndex = (DeviceParams->Type&SUBSTRUCT_MASK)>>SUBSTRUCT_SHIFT;
      stbase = SubStruct[StructIndex].sub_struct + ((bit32u)DeviceParams->Offset);
      sub_struct_size = SubStruct[StructIndex].sub_struct_size;
      sub_index_size1 = SubStruct[StructIndex].sub_index_size1;
      sub_index_size2 = SubStruct[StructIndex].sub_index_size2;

      for (i=0; i<sub_index_size1; i++)
        {
         for (j=0; j<sub_index_size2; j++, stbase += sub_struct_size)
           {
            /* skip this parameter if the integer value is 0 */
            if ((DeviceParams->Type&ISSTRING) == 0)
              if (*((bit32u*)(((bit32u)stbase))) == 0)
                 continue;

            if (sub_index_size1 > 1)
              {
               if (sub_index_size2 > 1)
                 OsFunc->Printf("%-18s[%2d][%2d] ",DeviceParams->pszKey,i,j);
                else
                 OsFunc->Printf("%-18s[%2d]     ",DeviceParams->pszKey,i);
              }
             else
              OsFunc->Printf("%-18s         ",DeviceParams->pszKey);

            if (DeviceParams->Type&ISSTRING)
              OsFunc->Printf("%-10s",((char*)(((bit32u)stbase))));
             else
              OsFunc->Printf("0x%08x",*((bit32u*)(((bit32u)stbase))));
            if (line++)
              {
               OsFunc->Printf("\n");
               line=0;
              }
             else
              OsFunc->Printf("  ");
           }
        }
      DeviceParams++;
     }

   if (line)
   OsFunc->Printf("\n");

   return(0);
  }
#endif

int cpswHalCommonParamsGetEnv(HAL_DEVICE *HalDev,OS_FUNCTIONS *OsFunc, OS_DEVICE *OsDev, const DEVICE_PARAMS *DeviceParams,SUB_STRUCT *SubStructs)
  {
   int rc=0;

   while(DeviceParams->pszKey)
     {
      if (DeviceParams->Type&ISENV)
        {
         /* get OS environment variable, then set in HAL */
         if (DeviceParams->Type&ISSTRING)
           {
            char **Str=0;
            rc = osPullRef(DeviceParams->pszKey, Str);
            if(rc)
              return (rc);

            rc = cpswHalCommonParamsControl(HalDev,OsFunc,DeviceParams,SubStructs,DeviceParams->pszKey,haSETbySTR,*Str);
            if(rc)
              return (rc);
           }
         else
           {
            bit32u Value=0;
            rc = osPullRef(DeviceParams->pszKey, &Value);
            if(rc)
              return (rc);

            rc = cpswHalCommonParamsControl(HalDev,OsFunc,DeviceParams,SubStructs,DeviceParams->pszKey,haSETbyREF,&Value);
            if(rc)
              return (rc);
           }
        }
     DeviceParams++;
     }
   return (EC_NO_ERRORS);
  }

int cpswHalCommonParamsGetConfig(HAL_DEVICE *HalDev,OS_FUNCTIONS *OsFunc, const CONFIG_PARAMS *cDeviceParams,const DEVICE_PARAMS *DeviceParams,SUB_STRUCT *SubStructs,int *inst)
  {
   char *pkey,*pkeyparm;
   bit32u keyerrorval,sts,more,loops,inst_index=0,ResetBase=0;
   char *handle=NULL;
   char *cp;

   handle=0;
   while(cDeviceParams->key)
     {
      pkey= (char *)cDeviceParams->key;
      pkeyparm= (char *)cDeviceParams->keyparm;
      keyerrorval=cDeviceParams->errorval;
      cDeviceParams++;
      if (!pkeyparm)
        {
         /* release previously used DeviceInfo handle */
         if (handle != NULL)
           OsFunc->DeviceFindInfoRelease(handle);

         sts=OsFunc->DeviceFindInfo(inst[inst_index++],pkey,&handle);
         if ((sts)&&(keyerrorval)) return(keyerrorval);
        }
       else
        {
         if (pkeyparm != hkResetBase)
           {
            /* fix this to use two indices */
            loops = cDeviceParams->index_size;
            more = 0;

            sts=OsFunc->DeviceFindParmValue(handle, pkey, &cp);

            if (sts == 0)
              {
               /* Uses haSETbySTR because all params from configuration store are strings */
               sts=cpswHalCommonParamsControl(HalDev,OsFunc,DeviceParams,SubStructs,pkeyparm,haSETbySTR,cp);
               if ((sts)&&(keyerrorval)) return(keyerrorval);
              }

            if (loops)
              {
               sts=cpswHalCommonParamsGetConfigArray(HalDev,OsFunc,DeviceParams,SubStructs,handle,pkey,pkeyparm,loops,keyerrorval,more);
               if ((sts)&&(keyerrorval)) return(keyerrorval);
              }

            /* release value string */
            OsFunc->DeviceFindParmValueRelease(cp);
           }
          else
           {
            /* Special code to extract reset base */
            sts=ResetBaseGet(OsFunc, &ResetBase, handle);
            if (!sts)
              {
               cpswHalCommonParamsControl(HalDev,OsFunc,DeviceParams,SubStructs,hkResetBase,haSETbyVAL,(bit32u *)ResetBase);
              }
           }
        }
     }
   return(0);
  }

int cpswHalCommonParamsGetConfigArray(HAL_DEVICE *HalDev,OS_FUNCTIONS *OsFunc, const DEVICE_PARAMS *DeviceParams,SUB_STRUCT *SubStructs,char *handle,char *pSKey,char *pWKey,int loops,bit32u err,int more)
  {
   char NewControlKey[40],NewSearchKey[40];
   int i,sts;
   char *RetData;

   if (loops)
     for(i=0;i<loops;i++)
       {
        cpswHalCommonstrcpy(NewSearchKey,pSKey);
        cpswHalCommonstrcat(NewSearchKey,".");
        cpswHalCommoncatdnum(NewSearchKey,i);
        cpswHalCommonstrcpy(NewControlKey,pWKey);
        cpswHalCommonstrcat(NewControlKey,".");
        cpswHalCommoncatdnum(NewControlKey,i);
        sts=OsFunc->DeviceFindParmValue(handle, NewSearchKey, &RetData);
        if (sts == 0)
          {
           sts=cpswHalCommonParamsControl(HalDev,OsFunc,DeviceParams,SubStructs,NewControlKey,haSETbySTR,RetData);
           if ((sts)&&(err)) return(err);
          }
        if (more)
          {
           sts=cpswHalCommonParamsGetConfigArray(HalDev,OsFunc,DeviceParams,SubStructs,handle,NewSearchKey,NewControlKey,more,err,0);
           if ((sts)&&(err)) return(err);
          }
        /* release used RetData value string */
        OsFunc->DeviceFindParmValueRelease(RetData);
       }
   return(0);
  }

int cpswHalCommonParamsCheckAllParams(HAL_DEVICE *HalDev,const DEVICE_PARAMS *DeviceParams, SUB_STRUCT *SubStruct)
  {
   bit32u val,hi,lo, StructIndex;
   void *stbase;

   while(DeviceParams->pszKey)
     {
      /* only check parameters which are not pass through */
      if (!(DeviceParams->Type & PASS_THROUGH))
        {
         /* probably need to add a check here to only check parameter values,
            not including statistics values or soft stats */
         if (DeviceParams->Type & WRITABLE)
           {
            StructIndex = (DeviceParams->Type&SUBSTRUCT_MASK)>>SUBSTRUCT_SHIFT;
            stbase = SubStruct[StructIndex].sub_struct;

            if (DeviceParams->Type&ISSTRING)
              val=cpswHalCommonstrlen(((char*)(((bit32u)stbase)+((bit32u)DeviceParams->Offset))));
             else
              val=*((bit32u*)(((bit32u)stbase)+((bit32u)DeviceParams->Offset)));
            hi=DeviceParams->HighValue;
            lo=DeviceParams->LowValue;

            if (!((val>=lo)&&(val<=hi)))
              {
               return(EC_VAL_OUT_OF_RANGE);
              }
           }
        }
      DeviceParams++;
     }

   return(0);
  }

/* we don't need to pass HalDev if we include it in the structure array */
int cpswHalCommonParamsInit(HAL_DEVICE *HalDev,const DEVICE_PARAMS *DeviceParams, SUB_STRUCT *SubStruct)
  {
   void *stbase;
   int StructIndex,sub_index_size1,sub_index_size2,sub_struct_size,i,j;

   while(DeviceParams->pszKey)
     {
      if (DeviceParams->Type&DO_INIT)
        {
         StructIndex = (DeviceParams->Type&SUBSTRUCT_MASK)>>SUBSTRUCT_SHIFT;
         stbase = SubStruct[StructIndex].sub_struct + ((bit32u)DeviceParams->Offset);
         sub_struct_size = SubStruct[StructIndex].sub_struct_size;
         sub_index_size1 = SubStruct[StructIndex].sub_index_size1;
         sub_index_size2 = SubStruct[StructIndex].sub_index_size2;

         for (i=0; i<sub_index_size1; i++)
           {
            for (j=0; j<sub_index_size2; j++)
              {
               if (DeviceParams->Type&ISSTRING)
                 {
                  if (cpswHalCommonstrlen((char *)DeviceParams->Default)<DeviceParams->HighValue)
                    *(char**)(stbase) = (char*)DeviceParams->Default;
                 }
               else
                  *((bit32u*)(((bit32u)stbase)))=DeviceParams->Default;

               /* only update the base if continuing this inner loop */
               if(j < (sub_index_size2 - 1))
                 ((bit32u)stbase) += sub_struct_size;
              }
            ((bit32u)stbase) += sub_struct_size;
           }
        }
      DeviceParams++;
     }
   return(0);
  }

/**
 *  @ingroup Params_Functions
 *
 *  @param  Action        Indicates the action to perform on the parameter (haGETbyREF, haSETbyREF, haSETbyVAL)
 *  @param  ParmValue         (Void *) Output parameter for GET actions, input parameter for SET actions.
 *
 *  This functions returns the 'value' of ParmValue is Action is Set
 *
 *  @return Value of ParmValue
 *
 */
bit32u cpswHalCommonParamsValueSet(HAL_CONTROL_ACTION Action, void *ParmValue)
{
    bit32u Value;
    switch(Action)
    {
        case haSETbyREF:
             Value = *(unsigned int *) ParmValue;
        break;
        case haSETbyVAL:
             Value = (bit32u) ParmValue;
        break;
        default:
             Value = 0xFFFFFFFF;
        break;
    }
    return(Value);
}
