/*
 *	Handle incoming frames
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_input.c,v 1.1.1.1.2.7 2003/07/26 18:49:52 qle Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/if_bridge.h>
#include <linux/netfilter_bridge.h>
#include "br_private.h"

#include <asm/uaccess.h>

unsigned char bridge_ula[6] = { 0x01, 0x80, 0xc2, 0x00, 0x00, 0x00 };

static int br_pass_frame_up_finish(struct sk_buff *skb)
{
	netif_rx(skb);

	return 0;
}

static void br_pass_frame_up(struct net_bridge *br, struct sk_buff *skb)
{
	struct net_device *indev;

	br->statistics.rx_packets++;
	br->statistics.rx_bytes += skb->len;

	indev = skb->dev;
	skb->dev = &br->dev;
	skb->pkt_type = PACKET_HOST;
	skb_push(skb, ETH_HLEN);
	skb->protocol = eth_type_trans(skb, &br->dev);

	
	NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_IN, skb, indev, NULL,
			br_pass_frame_up_finish);
}


static void __br_handle_frame_orig(struct sk_buff *skb)
{
	struct net_bridge *br;
	unsigned char *dest;
	struct net_bridge_fdb_entry *dst;
	struct net_bridge_port *p;
	int passedup;


	dest = skb->mac.ethernet->h_dest;
	
	p = skb->dev->br_port;
	br = p->br;
	passedup = 0;

	if (!(br->dev.flags & IFF_UP) ||
	    p->state == BR_STATE_DISABLED)
		goto freeandout;

	if (br->dev.flags & IFF_PROMISC) {
		struct sk_buff *skb2;

		skb2 = skb_clone(skb, GFP_ATOMIC);
		if (skb2) {
			passedup = 1;
			br_pass_frame_up(br, skb2);
		}
	}

	if (skb->mac.ethernet->h_source[0] & 1)
		goto freeandout;


	if (!passedup &&
	    (dest[0] & 1) &&
	    (br->dev.flags & IFF_ALLMULTI || br->dev.mc_list != NULL)) {
		struct sk_buff *skb2;

		skb2 = skb_clone(skb, GFP_ATOMIC);
		if (skb2) {
			passedup = 1;
			br_pass_frame_up(br, skb2);
		}
	}

	if (br->stp_enabled &&
	    !memcmp(dest, bridge_ula, 5) &&
	    !(dest[5] & 0xF0))
		goto handle_special_frame;

	if (p->state == BR_STATE_LEARNING ||
	    p->state == BR_STATE_FORWARDING)
		br_fdb_insert(br, p, skb->mac.ethernet->h_source, 0);

	if (p->state != BR_STATE_FORWARDING)
		goto freeandout;

	if (dest[0] & 1) {
#ifdef AVM_NO_FLOOD_OF_OWN_MACS
		/* don't flood forward if it comes from one of our MACs */
		unsigned char *src = skb->mac.ethernet->h_source;
		struct net_bridge_port *p;

		for (p = br->port_list; p != NULL; p = p->next) {
			if (!memcmp(src, p->dev->dev_addr, ETH_ALEN)) break;
		}
		if (!p) {
			br_flood_forward(br, skb, 1);
		} else {
/* printk(KERN_INFO "bridge: received non-unicast from own MAC *not* forwarded\n"); */
		}
#else
		br_flood_forward(br, skb, 1);
#endif
		if (!passedup)
			br_pass_frame_up(br, skb);
		else
			kfree_skb(skb);
		return;
	}

	dst = br_fdb_get(br, dest);

	if (dst != NULL && dst->is_local) {
		if (!passedup)
			br_pass_frame_up(br, skb);
		else
			kfree_skb(skb);
		br_fdb_put(dst);
		return;
	}

	if (dst != NULL) {
		br_forward(dst->dst, skb);
		br_fdb_put(dst);
		return;
	}

	br_flood_forward(br, skb, 0);
	return;

 handle_special_frame:
	if (!dest[5]) {
		br_stp_handle_bpdu(skb);
		return;
	}

 freeandout:
	kfree_skb(skb);
}

static void __br_handle_frame(struct sk_buff *skb)
{
#if defined (CONFIG_MIPS_AVALANCHE_FAST_BRIDGE)

	struct net_bridge *br;
	unsigned char *dest;
	struct net_bridge_fdb_entry *dst;
	struct net_bridge_port *p;
	dest = skb->mac.ethernet->h_dest;
	
	p = skb->dev->br_port;
	br = p->br;
	
	if((br->dev.flags &(IFF_UP)) && 
	   !((br->dev.flags &IFF_PROMISC)|| br->br_filter_active || (dest[0] & 1) ))  
	{
	        if (p->state == BR_STATE_FORWARDING)
		{
			dst = br_fdb_get(br, dest);

			if (dst != NULL)  
			{
				br_fdb_insert(br, p, skb->mac.ethernet->h_source, 0);
				if(!dst->is_local)
				{
		    		    br_forward(dst->dst, skb);
		    		    br_fdb_put(dst);
		    		    return;
				}
				else
				{
				    br_pass_frame_up(br, skb);
		    		    br_fdb_put(dst);
				    return;				 
				}
			}

		}
	}
#endif 
	__br_handle_frame_orig(skb);
}

static int br_handle_frame_finish(struct sk_buff *skb)
{
	struct net_bridge *br;

	br = skb->dev->br_port->br;
	read_lock(&br->lock);

#ifdef AVM_DELAYQUEUE
	if (br->tokens_per_tick && atomic_add_negative(-1, &br->tokens)) {
		/* queue or drop the packet */
		atomic_inc(&br->tokens);
#ifdef AVM_DELAYQUEUE_STAT
		if (sbk_queue_empty(&br->delayqueue)) { br->delayqueue_stat_suspends++; }
#endif
		if (skb_queue_len(&br->delayqueue) < br->delayqueue_max) {
			skb_queue_tail(&br->delayqueue, skb);
		} else {
			/* drop the packet */
#ifdef AVM_DELAYQUEUE_STAT
			br->delayqueue_stat_drops++;
#endif
			kfree_skb(skb);
		}
	} else {
		__br_handle_frame(skb);
	}
#ifdef AVM_DELAYQUEUE_STAT
	if (br->tokens_per_tick && (jiffies - br->delayqueue_stat_last_jiffies) >= 10*HZ/*more than 10 sec.*/) {
printk(KERN_INFO "bridge: br=0x%x suspends=%u tokenmax=%u tokens=%d drops=%u\n", br, br->delayqueue_stat_suspends, br->delayqueue_stat_tokenmax, atomic_read(br->tokens),
	  br->delayqueue_stat_drops);
		br->delayqueue_stat_last_jiffies = jiffies;
	}
#endif

#else
		__br_handle_frame(skb);
#endif

	read_unlock(&br->lock);
	return 0;
}

#ifdef AVM_DELAYQUEUE
void br_delayqueue_handle_frames(struct net_bridge *br)
{
	struct sk_buff *skb;

	read_lock(&br->lock);

	while(skb = skb_dequeue(&br->delayqueue)) {
		if (br->tokens_per_tick && atomic_add_negative(-1, &br->tokens)) {
			/* no more tokens to handle this frame - undo side effects on queue and tokens */
			atomic_inc(&br->tokens);
			skb_queue_head(&br->delayqueue, skb);
			break; // while
		}
		__br_handle_frame(skb);
	}
	read_unlock(&br->lock);
}

void br_delayqueue_discard_frames(struct net_bridge *br, struct net_device *dev)
{
	struct sk_buff *skb;
	int len;

	read_lock(&br->lock);

	len = skb_queue_len(&br->delayqueue);
	while(len-- && (skb = skb_dequeue(&br->delayqueue))) {
		if (skb->dev == dev) {
			kfree_skb(skb);
		} else {
			skb_queue_tail(&br->delayqueue, skb);
		}
	}
	read_unlock(&br->lock);
}
#endif /* AVM_DELAYQUEUE */

void br_handle_frame(struct sk_buff *skb)
{
	NF_HOOK(PF_BRIDGE, NF_BR_PRE_ROUTING, skb, skb->dev, NULL,
			br_handle_frame_finish);
}

